// Copyright (c) 1999, A.H. van den Boogaard

#include "elastic.h"
#include "myexcept.h"

using FeaTure::Vector;
using FeaTure::Matrix;
using FeaTure::CollectArray;
using FeaTure::StressType;
using FeaTure::uni;
using FeaTure::pss;
using FeaTure::psn;
using FeaTure::axi;
using FeaTure::gen;
using FeaTure::Material_Point;
using FeaTure::Structural_Point;
using FeaTure::Structural_type;

using std::cerr;
using std::endl;
using std::sqrt;

Elastic_Point::Elastic_Point( Elastic_type *mat, StressType st )
          :Structural_Point(st),parameters(mat)
{
   int stress_size = StressSize( st );

   strain.resize( stress_size );
   strain.clear();
   new_strain.resize( stress_size );
   new_strain.clear();
}

Elastic_Point::~Elastic_Point(){}

Matrix Elastic_Point::get_D_mat() const
{
   double Young, Poisson;
   int    stress_size;

   Young = parameters->get_Young();
   Poisson = parameters->get_Poisson();
   stress_size = StressSize( stress_type );

   return iso_lin_elas_D_mat( Young, Poisson, stress_size, stress_type );
}

Matrix Elastic_Point::get_xp_D_mat() const
{
   double Young, Poisson;
   int    stress_size;

   Young = parameters->get_Young();
   Poisson = parameters->get_Poisson();
   stress_size = ExpandedStressSize( stress_type );

   return iso_lin_elas_D_mat( Young, Poisson, stress_size, stress_type );
}

Vector Elastic_Point::get_xp_stress() const
{
   if ( stress_type == uni || stress_type == pss )
      return expand( get_stress() );
   else
      return get_xp_D_mat()*expand(new_strain);
}

Vector Elastic_Point::get_stress() const
{
   return get_D_mat()*new_strain;
}

void Elastic_Point::set_strain_increment( const Vector& strain_increment )
{
   new_strain = strain + strain_increment;
}

void Elastic_Point::finish_increment()
{
   strain = new_strain;
}

void Elastic_Point::GetRestartData( CollectArray &data ) const
{
   data.add( strain );
}

void Elastic_Point::PutRestartData( const CollectArray &data )
{
   data.extract( strain );
   new_strain = strain;
}

Elastic_Point_xp::Elastic_Point_xp( Elastic_type *mat, StressType st )
                                  :Elastic_Point(mat,st)
{
   int stress_size = ExpandedStressSize( st );

   strain.resize( stress_size );
   strain.clear();
   new_strain.resize( stress_size );
   new_strain.clear();

   strain_bar.resize( strain_bar_size() );
   prev_dstrain.resize( StressSize( st ) );
   strain_bar.clear();
   prev_dstrain.clear();
}

Elastic_Point_xp::~Elastic_Point_xp(){}

Matrix Elastic_Point_xp::get_D_mat() const
{
   double Young, Poisson;
   int    xstrsize;

   Young = parameters->get_Young();
   Poisson = parameters->get_Poisson();
   xstrsize = ExpandedStressSize( stress_type );

   return condense( iso_lin_elas_D_mat( Young, Poisson, xstrsize, stress_type ) );
}

Vector Elastic_Point_xp::get_stress() const
{
   return get_D_mat()*condense_vector(new_strain);
}

Vector Elastic_Point_xp::get_xp_stress() const
{
   return ( get_xp_D_mat()*new_strain );
}

Matrix Elastic_Point_xp::get_xp_D_mat() const
{
   double Young, Poisson;
   int    xstrsize;

   Young = parameters->get_Young();
   Poisson = parameters->get_Poisson();
   xstrsize = ExpandedStressSize( stress_type );

   return ( iso_lin_elas_D_mat( Young, Poisson, xstrsize, stress_type ) );
}

void Elastic_Point_xp::set_strain_increment( const Vector& strain_increment )
{
   Vector dstrain( strain_increment );
   Vector ddstrain = dstrain - prev_dstrain;

   prev_dstrain = dstrain;

   ExpandEps( dstrain, ddstrain, strain_bar );
   Elastic_Point::set_strain_increment( dstrain );
}

void Elastic_Point_xp::finish_increment()
{
   Elastic_Point::finish_increment();
   strain_bar.clear();
   prev_dstrain.clear();
}

ThermoElastic_Point::ThermoElastic_Point( ThermoElastic_type *mat, StressType st )
                               :Elastic_Point( mat, st ),LinearThermal_Point( mat, st )
{
   thermal_strain_old = Delta_eps_th = 0.0;
}

ThermoElastic_Point::~ThermoElastic_Point(){}

void ThermoElastic_Point::set_strain_increment( const Vector& strain_increment )
{
   Vector el_strain_incr(strain_increment);
   double DepsTh = Delta_eps_th;

   el_strain_incr(1) -= DepsTh;

   switch ( stress_type )
   {
      case axi:
      case gen: el_strain_incr(3) -= DepsTh;
      case pss:
      case psn: el_strain_incr(2) -= DepsTh;
      case uni: break;
   }

   Elastic_Point::set_strain_increment( el_strain_incr );
}

void ThermoElastic_Point::set_potential_increment( double Tinc )
{
   Delta_eps_th = Tinc * get_expansion_coefficient();

   LinearThermal_Point::set_potential_increment( Tinc );
   return;
}

void ThermoElastic_Point::finish_increment()
{
   thermal_strain_old += Delta_eps_th;
   Delta_eps_th = 0.0;
   Elastic_Point::finish_increment();
   LinearThermal_Point::finish_increment();
}

double ThermoElastic_Point::get_expansion_coefficient() const
{
#ifndef NOVARIABLERETURN
   return dynamic_cast<ThermoElastic_type *>
                     (LinearThermal_Point::parameters)->get_alpha();
#else
   return ((ThermoElastic_type *)LinearThermal_Point::parameters)->get_alpha();
#endif
}

Vector ThermoElastic_Point::get_xp_stress() const
{
   if ( stress_type == uni || stress_type == pss )
      return expand( Elastic_Point::get_stress() );
   else
   {
      Vector xstrain(expand(new_strain));
      if ( stress_type == psn )
         xstrain(3) -= ( thermal_strain_old + Delta_eps_th );

      return get_xp_D_mat()*xstrain;
   }
}

Vector ThermoElastic_Point::get_stress() const
{
   if ( stress_type == psn )
      return condense_vector( get_xp_stress() );
   else
      return Elastic_Point::get_stress();
}

void ThermoElastic_Point::GetRestartData( CollectArray &data ) const
{
   Elastic_Point::GetRestartData( data );
   LinearThermal_Point::GetRestartData( data );
   data.add( thermal_strain_old );
}

void ThermoElastic_Point::PutRestartData( const CollectArray &data )
{
   Elastic_Point::PutRestartData( data );
   LinearThermal_Point::PutRestartData( data );
   thermal_strain_old = data.extract();
}

NonlinearElastic_Point::NonlinearElastic_Point( NonlinearElastic_type *mat, StressType st )
                  :Structural_Point(st),parameters(mat)
{
   if ( st != uni )
      throw featured_error( "only uniaxial nonlinear elasticity is allowed" );

   int stress_size = StressSize( st );

   strain.resize( stress_size );
   strain.clear();
   new_strain.resize( stress_size );
   new_strain.clear();
}

NonlinearElastic_Point::~NonlinearElastic_Point(){}

Matrix NonlinearElastic_Point::get_D_mat() const
{
   double Young;

   Young = parameters->get_Young();

   Matrix D(1,1);
   D(1,1) = Young / sqrt( std::fabs(new_strain(1)) + 1 );
   return D;
}

Matrix NonlinearElastic_Point::get_xp_D_mat() const
{
   throw featured_error( "no contraction data available for nonlinear elasticity" );
   Matrix Dummy(1,1,0.0);
   return Dummy; // some compilers give a warning if no return statement is given
}

Vector NonlinearElastic_Point::get_stress() const
{
   double Young, sigma;
   Vector stress(1);

   Young = parameters->get_Young();

   sigma = 2*Young*( sqrt( std::fabs(new_strain(1))+1 ) - 1 );

   if ( new_strain(1) < 0 ) sigma = -sigma;

   stress(1) = sigma;

   return stress;
}

void NonlinearElastic_Point::set_strain_increment( const Vector& strain_increment )
{
   new_strain = strain + strain_increment;
}

void NonlinearElastic_Point::finish_increment()
{
   strain = new_strain;
}

void NonlinearElastic_Point::GetRestartData( CollectArray &data ) const
{
   data.add( strain );
}

void NonlinearElastic_Point::PutRestartData( const CollectArray &data )
{
   data.extract( strain );
   new_strain = strain;
}

Elastic_type::Elastic_type( int matnr ):Structural_type()
{
   nr = matnr;
   Young = 0;
   Poisson = 0;
}

Elastic_type::~Elastic_type(){}

#ifdef NOVARIABLERETURN
Material_Point* Elastic_type::new_Material_Point( StressType st )
#else
Elastic_Point* Elastic_type::new_Material_Point( StressType st )
#endif
{
   return new Elastic_Point(this,st);
}

int Elastic_type::set_Young( double young )
{
   Young = young;

   if ( Young <= 0.0 )
   {
      cerr << "Youngs modulus <= 0; this is not allowed" << endl;
      return 1;
   }
   return 0;
}

double Elastic_type::get_Young() const
{
   return Young;
}

int Elastic_type::set_Poisson( double poisson )
{
   Poisson = poisson;
   if ( Poisson < 0 || Poisson > 0.5 )
   {
      cerr << "Poisson's ratio must be between 0 and 0.5; input error" << endl;
      return 1;
   }
   return 0;
}

double Elastic_type::get_Poisson() const
{
   return Poisson;
}

Elastic_type_xp::Elastic_type_xp( int matnr ):Elastic_type(matnr){}

Elastic_type_xp::~Elastic_type_xp(){}

#ifdef NOVARIABLERETURN
Material_Point* Elastic_type_xp::new_Material_Point( StressType st )
#else
Elastic_Point_xp* Elastic_type_xp::new_Material_Point( StressType st )
#endif
{
   return new Elastic_Point_xp(this,st);
}

ThermoElastic_type::ThermoElastic_type( int matnr )
                   :Elastic_type( matnr ),LinearThermal_type(matnr)
{}

ThermoElastic_type::~ThermoElastic_type(){}

#ifdef NOVARIABLERETURN
Material_Point* ThermoElastic_type::new_Material_Point( StressType st )
#else
ThermoElastic_Point* ThermoElastic_type::new_Material_Point( StressType st )
#endif
{
   return new ThermoElastic_Point(this,st);
}

int ThermoElastic_type::set_alpha( double al )
{
   alpha = al;

   if ( alpha <= 0.0 )
   {
      cerr << "expansion coefficient <= 0; this is not allowed" << endl;
      return 1;
   }
   return 0;
}

NonlinearElastic_type::NonlinearElastic_type( int matnr ):Structural_type()
{
   Young = 0;
   nr = matnr;
}

NonlinearElastic_type::~NonlinearElastic_type(){}

#ifdef NOVARIABLERETURN
Material_Point* NonlinearElastic_type::new_Material_Point( StressType st )
#else
NonlinearElastic_Point* NonlinearElastic_type::new_Material_Point( StressType st )
#endif
{
   return new NonlinearElastic_Point(this,st);
}

int NonlinearElastic_type::set_Young( double young )
{
   Young = young;

   if ( Young <= 0.0 )
   {
      cerr << "Youngs modulus <= 0; this is not allowed" << endl;
      return 1;
   }
   return 0;
}

double NonlinearElastic_type::get_Young() const
{
   return Young;
}
