// Copyright (c) 1999,2005, A.H. van den Boogaard

#include "domain.cpp"
#include "control.cpp"
#include "ftparser.h"

#include "tlelemen.h"
#include "syntax.h"
#include "myexcept.h"

#include <cstring>
#include <cctype>

#define CONT_CHAR '\\'

using FeaTure::Control;
using FeaTure::Material_type;
using FeaTure::Geometry;
using FeaTure::Dof;
using FeaTure::LoadSet;
using FeaTure::axi;
using FeaTure::psn;
using FeaTure::pss;
using FeaTure::gen;

using std::cout;
using std::cerr;
using std::endl;
using std::tolower;
using std::isdigit;
using std::strpbrk;
using std::sscanf;

input_data::input_data()
{
   in_ptr = NULL;
   line_nr = 0;
   separators = " \t\n\0";
   separators += CONT_CHAR;
   field_start = field_end = 0;
}

void input_data::interprete_line( MyDomain *model, Control<MyDomain> **control )
{
   switch ( get_key() )
   {
      case TITLE     : read_title( model );          break;
      case NODES     : read_nodes( model );          break;
      case ELEMENTS  : read_elements( model );       break;
      case SUPPRESS  : read_suppress( model );       break;
      case MATERIAL  : read_material( model );       break;
      case GEOMETRY  : read_geometry( model );       break;
      case NONLINEAR : read_nonlinear( control );    break;
      case LINEAR    : read_linear( control );       break;
      case EIGEN     : read_eigen( control );        break;
      case LOADEF    : read_load_definition( model );break;
      case POST      : read_post( *control );        break;
      case END       : cout << "END\n";
      case NOKEY     : cout << "no_key\n";
      default        : cout << "error in switch(key)\n";
                       set_interpreted();
   }
}

input_data::input_data( std::string file_name )
{
   set_file_name( file_name );
   in_ptr = NULL;
   line_nr = 0;
}

input_data::~input_data()
{
   if ( in_ptr != NULL ) fclose( in_ptr );
}

int input_data::open_file()
{
   line_nr = 0;
   error_in_file = 0;

   if ( filename[0] == '\0' )
   {
      cerr << "Name of input file not initialized yet" << endl;
      return 1;
   }

   if ((in_ptr = std::fopen( filename.c_str(), "rt")) == NULL)
   {
      cerr << "Cannot open input file: " << filename << endl;
      return 1;
   }

   return 0;
}

void input_data::read_next_line()
{
   int test_key;

   if ( in_ptr == NULL )
     if ( open_file() )
     {
        key = END;
        interpreted = new_key = 1;
        error_in_file++;
        return;
     }

   if ( line_nr > 0 && error_in_line > 0 )
   {
      print_error_line();
      error_in_file += error_in_line;
   }

   nr_of_fields = 0;

   while ( nr_of_fields==0 && !eof() ) read_line();

   // if nr_of_fields == 0 at this point, we have reached the end of the file
   interpreted = ( nr_of_fields == 0 );
   if ( interpreted ) key = END;
   
   error_in_line = 0;
   current_field = 0;
   if ( nr_of_fields > 0 ) set_field(1);

   test_key = is_key();
   new_key  = ( test_key != NOKEY );
   if ( new_key ) key = test_key;
}

int input_data::get_key()
{
   return key;
}

void input_data::get_field( int field, double *item )
{
   int ret_val;

   set_field( field );
   ret_val = sscanf( current_word.c_str(), "%lf", item );

   if ( ret_val != 1 )
   {
      error_in_line++;
      cerr << "wrong type in field " << field << endl;
   }

   return;
}

void input_data::get_field( int field, int *item )
{
   int ret_val;

   set_field( field );
   ret_val = sscanf( current_word.c_str(), "%d", item );

   if ( ret_val != 1 )
   {
      error_in_line++;
      cerr << "wrong type in field " << field << endl;
   }

   return;
}

void input_data::get_field( int field, std::string &word )
{
   set_field( field );

   word = current_word;

   if ( word.empty() )
   {
      error_in_line++;
      cerr << "missing string in field " << field << endl;
   }

   return;
}

void input_data::get_field_lower_case( int field, std::string &word )
{
   set_field( field );

   std::string::const_iterator s=current_word.begin();

   word.resize(0);
   while ( s!=current_word.end() ) word += (char)tolower(*s++);

   if ( word.empty() )
   {
      error_in_line++;
      cerr << "missing string in field " << field << endl;
   }

   return;
}

void input_data::set_field( int field )
{
   if ( field > nr_of_fields )
   {
      cerr << "program error, attempt to read field " << field << " of "
           << nr_of_fields << " available fields" << endl;

      print_error_line();
      throw syntax_error();
   }

   if ( field < current_field || current_field < 1 )
   {
      current_field = 1;
      field_start = input_line.find_first_not_of( separators );
      field_end = input_line.find_first_of( separators, field_start+1 );
   }

   while ( current_field < field )
   {
      field_start = input_line.find_first_not_of( separators, field_end+1 );
      field_end   = input_line.find_first_of( separators, field_start+1 );
      current_field++;
   }

   if ( field_end == std::string::npos ) field_end = input_line.size();
   current_word = input_line.substr( field_start, field_end - field_start );
}

int input_data::line_is_interpreted()
{
   return ( interpreted );
}

int input_data::is_new_key()
{
   return ( new_key );
}

int input_data::eof()
{
   if ( in_ptr == NULL ) return 1;

   return ( feof( in_ptr ) );
}

void input_data::print_line()
{
   cout << input_line << endl;
}

void input_data::print_error_line()
{
   cerr << endl << "error in input line " << line_nr << ":\n";
   cerr << input_line << endl << endl;
}

void input_data::print_field( int field )
{
   set_field( field );
   cerr << current_word;
}

void input_data::set_interpreted()
{
   interpreted = 1;
}

int input_data::interpreted_all()
{
   if ( line_nr == 0 || line_is_interpreted() ) read_next_line();

   return ( key == END || ( eof() && line_is_interpreted() ) );
}

int input_data::read_title( MyDomain *model )
{
   std::string title;

   read_next_line();
   while ( !is_new_key() )
   {
      if ( title.empty() ) title = input_line;
      else
      {
         title += '\n';
         title += input_line;
      }
      read_next_line();
   }

   model->set_title( title );
   return (1);
}

int input_data::read_nodes( MyDomain* model )
{
   int   nr;
   double x, y, z;
   MyNode* node;

   read_next_line();

   while ( !is_new_key() )
   {
      get_field( 1, &nr );
      get_field( 2, &x );
      get_field( 3, &y );
      get_field( 4, &z );

      if ( !error_in_line )
      {
         node = new MyNode( nr, x, y, z );
         model->add_node( node );
      }

      read_next_line();
   }

   return (1);
}

int input_data::read_elements( MyDomain *model )
{
   int         elm_key, elm_nr, mat_nr, geom_nr, rqd_nodes;
   int         dataerror;
   Material_type *matptr;
   Geometry    *geomptr;
   Node_ptr    *node_array;

   read_next_line();

   while ( !is_new_key() )
   {
      elm_key = is_element_name();
      if ( elm_key != NOELM )
      {
         if ( nr_of_fields > 1 )
         {
            get_field( 2, &mat_nr );
            matptr = model->get_mat_ptr( mat_nr );
            if ( matptr == NULL )
            {
               cerr << "Error: material nr " << mat_nr <<
                      " must be defined before used" << endl;
               error_in_line++;
            }
         }
         else matptr = NULL;

         if ( nr_of_fields > 2 )
         {
            get_field( 3, &geom_nr );
            geomptr = model->get_geom_ptr( geom_nr );
            if ( geomptr == NULL )
            {
               cerr << "Error: geometry nr " << geom_nr <<
                      " must be defined before used" << endl;
               error_in_line++;
            }
         }
         else geomptr = NULL;

         switch ( elm_key )
         {
            case TRUSS   : rqd_nodes = Truss::reqd_nr_nodes()  ;
                           dataerror = Truss::check_data( matptr, geomptr );
                           break;
            case BEAM2D  : rqd_nodes = LinBeam2d::reqd_nr_nodes()  ;
                           dataerror = LinBeam2d::check_data( matptr, geomptr );
                           break;
            case TRI3PSN : rqd_nodes = Tri3psn::reqd_nr_nodes();
                           dataerror = Tri3psn::check_data( matptr, geomptr, psn );
                           break;
            case TRI3PSS : rqd_nodes = Tri3pss::reqd_nr_nodes();
                           dataerror = Tri3pss::check_data( matptr, geomptr, pss );
                           break;
            case TRI3AXI : rqd_nodes = Tri3axi::reqd_nr_nodes();
                           dataerror = Tri3axi::check_data( matptr, geomptr, axi );
                           break;
            case TRI3THERM:rqd_nodes = Tri3therm::reqd_nr_nodes();
                           dataerror = Tri3therm::check_data( matptr, geomptr );
                           break;
            case TRI3PSNTHERM:
                           rqd_nodes = Tri3psnTherm::reqd_nr_nodes();
                           dataerror = Tri3psnTherm::check_data( matptr, geomptr, psn );
                           break;
            case TRI6PSN : rqd_nodes = Tri6psn::reqd_nr_nodes();
                           dataerror = Tri6psn::check_data( matptr, geomptr, psn );
                           break;
            case TRI6PSS : rqd_nodes = Tri6pss::reqd_nr_nodes();
                           dataerror = Tri6pss::check_data( matptr, geomptr, pss );
                           break;
            case TRI6AXI : rqd_nodes = Tri6axi::reqd_nr_nodes();
                           dataerror = Tri6axi::check_data( matptr, geomptr, axi );
                           break;
            case TRI6PSNA: rqd_nodes = Tri6psn::reqd_nr_nodes();
                           dataerror = Tri6psn::check_data( matptr, geomptr, psn );
                           break;
            case TRI6PSSA: rqd_nodes = Tri6pss::reqd_nr_nodes();
                           dataerror = Tri6pss::check_data( matptr, geomptr, pss );
                           break;
            case TRI6AXIA: rqd_nodes = Tri6axi::reqd_nr_nodes();
                           dataerror = Tri6axi::check_data( matptr, geomptr, axi );
                           break;
            case QUAD4PSN: rqd_nodes = Quad4psn::reqd_nr_nodes();
                           dataerror = Quad4psn::check_data( matptr, geomptr, psn );
                           break;
            case QUAD4PSS: rqd_nodes = Quad4pss::reqd_nr_nodes();
                           dataerror = Quad4pss::check_data( matptr, geomptr, pss );
                           break;
            case QUAD4AXI: rqd_nodes = Quad4axi::reqd_nr_nodes();
                           dataerror = Quad4axi::check_data( matptr, geomptr, axi );
                           break;
            case QUAD4PSSTHERM:
                           rqd_nodes = Quad4pssTherm::reqd_nr_nodes();
                           dataerror = Quad4pssTherm::check_data( matptr, geomptr, pss );
                           break;
            case QUAD4PSNTHERM:
                           rqd_nodes = Quad4psnTherm::reqd_nr_nodes();
                           dataerror = Quad4psnTherm::check_data( matptr, geomptr, psn );
                           break;
            case TLQUAD4PSN:
                           rqd_nodes = TLQuad4psn::reqd_nr_nodes();
                           dataerror = TLQuad4psn::check_data( matptr, geomptr, psn );
                           break;
            case QUAD8PSN: rqd_nodes = Quad8psn::reqd_nr_nodes();
                           dataerror = Quad8psn::check_data( matptr, geomptr, psn );
                           break;
            case QUAD8PSS: rqd_nodes = Quad8pss::reqd_nr_nodes();
                           dataerror = Quad8pss::check_data( matptr, geomptr, pss );
                           break;
            case QUAD8AXI: rqd_nodes = Quad8axi::reqd_nr_nodes();
                           dataerror = Quad8axi::check_data( matptr, geomptr, axi );
                           break;
            case QUAD8PSNA:rqd_nodes = Quad8psn::reqd_nr_nodes();
                           dataerror = Quad8psn::check_data( matptr, geomptr, psn );
                           break;
            case QUAD8PSSA:rqd_nodes = Quad8pss::reqd_nr_nodes();
                           dataerror = Quad8pss::check_data( matptr, geomptr, pss );
                           break;
            case QUAD8AXIA:rqd_nodes = Quad8axi::reqd_nr_nodes();
                           dataerror = Quad8axi::check_data( matptr, geomptr, axi );
                           break;
            case TLQUAD8PSN:
                           rqd_nodes = TLQuad8psn::reqd_nr_nodes();
                           dataerror = TLQuad8psn::check_data( matptr, geomptr, psn );
                           break;
            case HEX8    : rqd_nodes = Hex8::reqd_nr_nodes();
                           dataerror = Hex8::check_data( matptr, geomptr, gen );
                           break;
            case MASS2D:   rqd_nodes = Mass2D::reqd_nr_nodes();
                           dataerror = Mass2D::check_data( matptr, geomptr );
                           break;
            case MASS3D:   rqd_nodes = Mass3D::reqd_nr_nodes();
                           dataerror = Mass3D::check_data( matptr, geomptr );
                           break;

            default      : rqd_nodes = 0; dataerror = 0;
         }

         error_in_line += dataerror;

         read_next_line();

         while ( isdigit( input_line[0] ) )
         {
            get_field( 1, &elm_nr );
            if ( !error_in_line )
            {
               node_array = get_node_array( model, rqd_nodes );
               if ( !error_in_line )
                  create_element( elm_key, model, matptr, geomptr, elm_nr, node_array );
            }
            read_next_line();
         }
      }
      else
      {
         error_in_line++;
         cerr << "unknown element name: ";
         print_field(1); cerr << endl;

         read_next_line();
         while ( isdigit( input_line[0] ) ) read_next_line();
      }
   }
   return (1);
}

FeaTure::Node** input_data::get_node_array( MyDomain *model, int rqd_nodes )
{
   Node_ptr    nodeptr, *node_array;
   int         node, node_nr;

   if ( rqd_nodes != nr_of_fields-1 )
   {
      error_in_line++;
      cerr << "wrong number of nodes " << (nr_of_fields-1) <<
              ", should be " << rqd_nodes << endl;
      return NULL;
   }

   node_array = new Node_ptr[rqd_nodes];

   for ( node=1; node <= rqd_nodes; node++ )
   {
      get_field( node+1, &node_nr );
      if ( error_in_line > 0 ) return NULL;

      nodeptr = model->get_node_ptr( node_nr );
      if ( nodeptr == NULL )
      {
         error_in_line++;
         cerr << "unknown nodenr " << node_nr << endl;
         return NULL;
      }
      node_array[node-1] = nodeptr;
   }

   return node_array;
}

void input_data::create_element( int elm_key, MyDomain *model, Material_type *matptr,
                                 Geometry *geomptr, int elm_nr, Node_ptr *node_array )
{
   FiniteElement* elmptr;

   switch ( elm_key )
   {
      case TRUSS   : elmptr = new Truss( elm_nr, node_array, matptr, geomptr );
                     break;
      case BEAM2D  : elmptr = new LinBeam2d( elm_nr, node_array, matptr, geomptr );
                     break;
      case TRI3PSN : elmptr = new Tri3psn( elm_nr, node_array, matptr, geomptr );
                     break;
      case TRI3PSS : elmptr = new Tri3pss( elm_nr, node_array, matptr, geomptr );
                     break;
      case TRI3AXI : elmptr = new Tri3axi( elm_nr, node_array, matptr, geomptr );
                     break;
      case TRI3THERM:elmptr = new Tri3therm( elm_nr, node_array, matptr, geomptr );
                     break;
      case TRI3PSNTHERM:
                     elmptr = new Tri3psnTherm( elm_nr, node_array, matptr, geomptr );
                     break;
      case TRI6PSN : elmptr = new Tri6psn( elm_nr, node_array, matptr, geomptr, false );
                     break;
      case TRI6PSS : elmptr = new Tri6pss( elm_nr, node_array, matptr, geomptr, false );
                     break;
      case TRI6AXI : elmptr = new Tri6axi( elm_nr, node_array, matptr, geomptr, false );
                     break;
      case TRI6PSNA: elmptr = new Tri6psn( elm_nr, node_array, matptr, geomptr, true );
                     break;
      case TRI6PSSA: elmptr = new Tri6pss( elm_nr, node_array, matptr, geomptr, true );
                     break;
      case TRI6AXIA: elmptr = new Tri6axi( elm_nr, node_array, matptr, geomptr, true );
                     break;
      case QUAD4PSN: elmptr = new Quad4psn( elm_nr, node_array, matptr, geomptr );
                     break;
      case QUAD4PSS: elmptr = new Quad4pss( elm_nr, node_array, matptr, geomptr );
                     break;
      case QUAD4AXI: elmptr = new Quad4axi( elm_nr, node_array, matptr, geomptr );
                     break;
      case QUAD4PSSTHERM:
                     elmptr = new Quad4pssTherm( elm_nr, node_array, matptr, geomptr );
                     break;
      case QUAD4PSNTHERM:
                     elmptr = new Quad4psnTherm( elm_nr, node_array, matptr, geomptr );
                     break;
      case TLQUAD4PSN:
                     elmptr = new TLQuad4psn( elm_nr, node_array, matptr, geomptr );
                     break;
      case QUAD8PSN: elmptr = new Quad8psn( elm_nr, node_array, matptr, geomptr, false );
                     break;
      case QUAD8PSS: elmptr = new Quad8pss( elm_nr, node_array, matptr, geomptr, false );
                     break;
      case QUAD8AXI: elmptr = new Quad8axi( elm_nr, node_array, matptr, geomptr, false );
                     break;
      case QUAD8PSNA:elmptr = new Quad8psn( elm_nr, node_array, matptr, geomptr, true );
                     break;
      case QUAD8PSSA:elmptr = new Quad8pss( elm_nr, node_array, matptr, geomptr, true );
                     break;
      case QUAD8AXIA:elmptr = new Quad8axi( elm_nr, node_array, matptr, geomptr, true );
                     break;
      case TLQUAD8PSN:
                     elmptr = new TLQuad8psn( elm_nr, node_array, matptr, geomptr );
                     break;
      case HEX8    : elmptr = new Hex8( elm_nr, node_array, matptr, geomptr );
                     break;
      case MASS2D:   elmptr = new Mass2D( elm_nr, node_array, matptr, geomptr );
                     break;
      case MASS3D:   elmptr = new Mass3D( elm_nr, node_array, matptr, geomptr );
                     break;

      default:       error_in_line++;
                     cerr << "unknown element in input_data::create_element()" << endl;
                     return;
   }

   model->add_element( elmptr );

   return;
}

int input_data::read_suppress( MyDomain *model )
{
   int      field, nr_of_pairs, node_nr;
   Dof::Type phys_type;
   Node_ptr node;

   struct PAIR
   {
      Dof::Type type;
      FeaTure::Direction dir;
   };
   PAIR  pairs[10];

   read_next_line();

   while ( !is_new_key() )
   {
      field = 1;
      set_field(field);
      while ( !isdigit( current_word[0] ) && field <= nr_of_fields )
      {
         if ( field > 10 )
         {
            cerr << "too many directions (must be less than 11)\n";
            error_in_line++;
            break;
         }

         if ( is_dof_type( current_word[0], &phys_type ) )
         {
            pairs[field-1].type = phys_type;
            if ( is_vector_dof( phys_type ) )
               get_direction( current_word.substr(1,current_word.size()-1),
                              &(pairs[field-1].dir) );
         }
         else
         {
            cerr << "unknown type of load" << endl;
            error_in_line++;
         }

         field++;
         set_field(field);
      }
      nr_of_pairs = field-1;

      for ( int i=field; i<=nr_of_fields; i++ )
      {
         get_field( i, &node_nr );
         if ( error_in_line ) break;
         node = model->get_node_ptr( node_nr );
         if ( node == NULL )
         {
            cerr << "unknown suppressed node number" << node_nr << endl;
            error_in_line++;
            continue;
         }

         for ( int dof = 0; dof<nr_of_pairs; dof++ )
            node->create_dof( Dof::fixed, phys_type, pairs[dof].dir );
      }

      read_next_line();
   }

   return 1;
}

int input_data::read_prescribed( MyDomain *model, LoadSet *load_set )
{
   int      field, nr_of_pairs, node_nr, nr_nodes;
   Dof::Type type;
   FeaTure::Nodal_Prescribed *nodal_load;
   double   magnitude;
   FeaTure::Direction dir;
   Node_ptr node;

   read_next_line();

   while ( !is_new_key() )
   {
      nodal_load = new FeaTure::Nodal_Prescribed;

      field = 1;
      set_field(field);
      while ( !isdigit( current_word[0] ) && field <= nr_of_fields )
      {
         if ( is_dof_type( current_word[0], &type ) )
         {
            get_field( field+1, &magnitude );
            set_field( field );

            if ( is_vector_dof( type ) )
            {
               get_direction( current_word.substr(1,current_word.size()-1), &dir);
               nodal_load->add_load( type, dir, magnitude );
            }
            else nodal_load->add_load( type, magnitude );
         }
         else
         {
            return 0;
         }
         field+=2;
         set_field(field);
      }
      nr_of_pairs = field-1;
      nr_nodes = nr_of_fields - nr_of_pairs;

      for ( int i=1; i<=nr_nodes; i++ )
      {
         get_field( i+nr_of_pairs, &node_nr );
         if ( error_in_line ) break;

         node = model->get_node_ptr( node_nr );
         if ( node == NULL )
         {
            cerr << "unknown prescribed node number" << node_nr << endl;
            error_in_line++;
            continue;
         }

         nodal_load->add_node( node );
      }

      load_set->add_load( nodal_load );

      read_next_line();
   }

   return 1;
}

int input_data::read_nodal_loads( MyDomain *model, LoadSet *load_set )
{
   int       field, nr_of_pairs, node_nr, nr_nodes;
   Dof::Load type;
   FeaTure::Direction dir;
   FeaTure::Nodal_Loads *nodal_load;
   double    magnitude;
   Node_ptr  node;

   read_next_line();

   while ( !is_new_key() )
   {
      nodal_load = new FeaTure::Nodal_Loads;
      field = 1;
      set_field(field);
      while ( !isdigit( current_word[0] ) && field <= nr_of_fields )
      {
         if ( is_load_type( current_word[0], &type ) )
         {
            get_field( field+1, &magnitude );
            set_field( field );
            if ( is_vector_load( type ) )
            {
               get_direction( current_word.substr(1,current_word.size()-1), &dir);
               nodal_load->add_load( type, dir, magnitude );
            }
            else nodal_load->add_load( type, magnitude );
         }
         else
         {
            return 0;
         }
         field += 2;
         set_field(field);
      }
      nr_of_pairs = field-1;
      nr_nodes = nr_of_fields - nr_of_pairs;

      for ( int i=1; i<=nr_nodes; i++ )
      {
         get_field( i+nr_of_pairs, &node_nr );
         if ( error_in_line ) break;
         node = model->get_node_ptr( node_nr );
         if ( node == NULL )
         {
            cerr << "unknown node number " << node_nr << " in nodal loads" << endl;
            error_in_line++;
            continue;
         }
         nodal_load->add_node( node );
      }
      load_set->add_load( nodal_load );

      read_next_line();
   }

   return 1;
}

int input_data::read_material( MyDomain *model )
{
   int      mat_nr;
   Material_type  *mat;

   if ( nr_of_fields > 1 ) get_field( 2, &mat_nr );
   else mat_nr = 1;

   read_next_line();

   std::string word;
   get_field_lower_case( 1, word );

   if      ( word == "linearelastic" )
   {
      read_next_line();
      mat = read_elastic_material( mat_nr );
   }
   else if ( word == "xplinearelastic" )
   {
      read_next_line();
      mat = read_xp_elastic_material( mat_nr );
   }
   else if ( word == "nonlinearelastic" )
   {
      read_next_line();
      mat = read_nonlinearelastic_material( mat_nr );
   }
   else if ( word == "vonmises" )
   {
      read_next_line();
      mat = read_vonmises_material( mat_nr );
   }
   else if ( word == "thermal" )
   {
      read_next_line();
      mat = read_thermal_material( mat_nr );
   }
   else if ( word == "thermoelastic" )
   {
      read_next_line();
      mat = read_thermoelastic_material( mat_nr );
   }
   else if ( word == "thermoplastic" )
   {
      read_next_line();
      mat = read_thermoplastic_material( mat_nr );
   }
   else if ( word == "mass" )
   {
      mat = read_mass_only_material( mat_nr );
   }
   else
      /* default linear elastic material */
      mat = read_elastic_material( mat_nr );

   model->add_material( mat );

   return 1;
}

Elastic_type_xp* input_data::read_xp_elastic_material( int mat_nr )
{
   Elastic_type_xp  *mat;
   double      value;

   mat = new Elastic_type_xp( mat_nr );

   while ( !is_new_key() )
   {
      std::string word;
      get_field_lower_case( 1, word );

      if ( word == "young" )
      {
         get_field( 2, &value ); error_in_line += mat->set_Young( value );
      }
      else if ( word == "poisson" )
      {
         get_field( 2, &value ); error_in_line += mat->set_Poisson( value );
      }
      else if ( word == "density" )
      {
         get_field( 2, &value ); error_in_line += mat->set_density( value );
      }
      else error_in_line++;

      read_next_line();
   }
   return mat;
}

Elastic_type* input_data::read_elastic_material( int mat_nr )
{
   Elastic_type  *mat;
   double      value;

   mat = new Elastic_type( mat_nr );

   while ( !is_new_key() )
   {
      std::string word;
      get_field_lower_case( 1, word );

      if ( word == "young" )
      {
         get_field( 2, &value ); error_in_line += mat->set_Young( value );
      }
      else if ( word == "poisson" )
      {
         get_field( 2, &value ); error_in_line += mat->set_Poisson( value );
      }
      else if ( word == "density" )
      {
         get_field( 2, &value ); error_in_line += mat->set_density( value );
      }
      else error_in_line++;

      read_next_line();
   }
   return mat;
}

NonlinearElastic_type* input_data::read_nonlinearelastic_material( int mat_nr )
{
   NonlinearElastic_type  *mat;
   double   value;

   mat = new NonlinearElastic_type( mat_nr );

   while ( !is_new_key() )
   {
      std::string word;
      get_field_lower_case( 1, word );

      if ( word == "young" )
      {
         get_field( 2, &value ); error_in_line += mat->set_Young( value );
      }
      else if ( word == "density" )
      {
         get_field( 2, &value ); error_in_line += mat->set_density( value );
      }
      else error_in_line++;

      read_next_line();
   }
   return mat;
}

VonMises_type* input_data::read_vonmises_material( int mat_nr )
{
   VonMises_type  *mat;
   double   value;

   mat = new VonMises_type( mat_nr );

   while ( !is_new_key() )
   {
      std::string word;
      get_field_lower_case( 1, word );

      if ( word == "young" )
      {
         get_field( 2, &value ); error_in_line += mat->set_Young( value );
      }
      else if ( word == "poisson" )
      {
         get_field( 2, &value ); error_in_line += mat->set_Poisson( value );
      }
      else if ( word == "density" )
      {
         get_field( 2, &value ); error_in_line += mat->set_density( value );
      }
      else if ( word == "nadai" )
      {
         get_field( 2, &value ); error_in_line += mat->set_nadai_C( value );
         get_field( 3, &value ); error_in_line += mat->set_nadai_n( value );
         get_field( 4, &value ); error_in_line += mat->set_yield( value );
      }
      else error_in_line++;

      read_next_line();
   }
   return mat;
}

LinearThermal_type* input_data::read_thermal_material( int mat_nr )
{
   LinearThermal_type  *mat;
   double   value;

   mat = new LinearThermal_type( mat_nr );

   while ( !is_new_key() )
   {
      std::string word;
      get_field_lower_case( 1, word );

      if ( word == "lambda" )
      {
         get_field( 2, &value ); error_in_line += mat->set_conductivity( value );
      }
      else if ( word == "capacity" )
      {
         get_field( 2, &value ); error_in_line += mat->set_capacity( value );
      }
      else if ( word == "density" )
      {
         get_field( 2, &value ); error_in_line += mat->set_density( value );
      }
      else error_in_line++;

      read_next_line();
   }
   return mat;
}

ThermoElastic_type* input_data::read_thermoelastic_material( int mat_nr )
{
   ThermoElastic_type  *mat;
   double   value;

   mat = new ThermoElastic_type( mat_nr );

   while ( !is_new_key() )
   {
      std::string word;
      get_field_lower_case( 1, word );

      if ( word == "young" )
      {
         get_field( 2, &value ); error_in_line += mat->set_Young( value );
      }
      else if ( word == "poisson" )
      {
         get_field( 2, &value ); error_in_line += mat->set_Poisson( value );
      }
      else if ( word == "lambda" )
      {
         get_field( 2, &value ); error_in_line += mat->set_conductivity( value );
      }
      else if ( word == "capacity" )
      {
         get_field( 2, &value ); error_in_line += mat->set_capacity( value );
      }
      else if ( word == "density" )
      {
         get_field( 2, &value ); error_in_line += mat->set_density( value );
      }
      else if ( word == "alpha" )
      {
         get_field( 2, &value ); error_in_line += mat->set_alpha( value );
      }
      else error_in_line++;

      read_next_line();
   }
   return mat;
}

ThermoPlastic_type* input_data::read_thermoplastic_material( int mat_nr )
{
   ThermoPlastic_type  *mat;
   double   value;

   mat = new ThermoPlastic_type( mat_nr );

   while ( !is_new_key() )
   {
      std::string word;
      get_field_lower_case( 1, word );

      if ( word == "young" )
      {
         get_field( 2, &value ); error_in_line += mat->set_Young( value );
      }
      else if ( word == "poisson" )
      {
         get_field( 2, &value ); error_in_line += mat->set_Poisson( value );
      }
      else if ( word == "lambda" )
      {
         get_field( 2, &value ); error_in_line += mat->set_conductivity( value );
      }
      else if ( word == "capacity" )
      {
         get_field( 2, &value ); error_in_line += mat->set_capacity( value );
      }
      else if ( word == "density" )
      {
         get_field( 2, &value ); error_in_line += mat->set_density( value );
      }
      else if ( word == "alpha" )
      {
         get_field( 2, &value ); error_in_line += mat->set_alpha( value );
      }
      else if ( word == "nadai" )
      {
         get_field( 2, &value ); error_in_line += mat->set_nadai_C( value );
         get_field( 3, &value ); error_in_line += mat->set_nadai_n( value );
         get_field( 4, &value ); error_in_line += mat->set_yield( value );
      }
      else error_in_line++;

      read_next_line();
   }
   return mat;
}

FeaTure::Mass_type* input_data::read_mass_only_material( int mat_nr )
{
   FeaTure::Mass_type  *mat;
   double   value;

   mat = new FeaTure::Mass_type( mat_nr );

   get_field( 2, &value );
   error_in_line += mat->set_mass( value );

   read_next_line();

   while ( !is_new_key() )
   {
      error_in_line++;
      read_next_line();
   }
   return mat;
}

int input_data::read_geometry( MyDomain *model )
{
   int      geom_nr;
   Geometry *geom;
   double   value;

   if ( nr_of_fields > 1 ) get_field( 2, &geom_nr );
   else geom_nr = 1;

   geom = new Geometry( geom_nr );

   read_next_line();

   while ( !is_new_key() )
   {
      std::string word;
      get_field_lower_case( 1, word );

      if ( word == "thick" )
      {
         get_field( 2, &value ); geom->set_Thick( value );
      }
      else if ( word == "crosse" )
      {
         get_field( 2, &value ); geom->set_Crosse( value );
      }
      else if ( word == "iz" )
      {
         get_field( 2, &value ); geom->set_Iz( value );
      }
      else if ( word == "iy" )
      {
         get_field( 2, &value ); geom->set_Iy( value );
      }
      else if ( word == "j" )
      {
         get_field( 2, &value ); geom->set_J( value );
      }
      else error_in_line++;

      read_next_line();
   }
   model->add_geometry( geom );

   return 1;
}

int input_data::read_linear( Control<MyDomain> **control )
{
   *control = new LinearControl;
   read_next_line();

   return 1;
}

int input_data::read_eigen( Control<MyDomain> **control )
{
   EigenControl *eigencontrol;
   eigencontrol = new EigenControl;

   read_next_line();

   int    ivalue;
   double dvalue;

   while ( !is_new_key() )
   {
      std::string word;
      get_field_lower_case( 1, word );

      if ( word == "frequencies" )
      {
         get_field( 2, &ivalue );
         eigencontrol->set_nr_frequencies( ivalue );
      }
      else if ( word == "shift" )
      {
         get_field( 2, &dvalue );
         eigencontrol->set_shift( dvalue );
      }
      else error_in_line++;

      read_next_line();
   }

   *control = eigencontrol;

   return 1;
}

int input_data::read_nonlinear( Control<MyDomain> **control )
{
   int integration_type=1;

   if ( nr_of_fields > 1 )
   {
      std::string word;
      get_field_lower_case( 2, word );

      if      ( word == "static"  ) integration_type = 1;
      else if ( word == "dynamic" ) integration_type = 2;
      else
      {
         error_in_line++;
         cerr << "*NONLINEAR can only be supplemented by STATIC or DYNAMIC" << endl;
      }
   }

   if ( integration_type == 1 )
      return read_nonlinear_static( control );
   else if ( integration_type == 2 )
      return read_nonlinear_dynamic( control );
   else
   {
      cerr << "unknown integration type, program error in input_data::read_nonlinear"
           << endl;
      return 0;
   }
}

int input_data::read_nonlinear_static( Control<MyDomain> **control )
{
   NonLinearStaticControl *NLScontrol = new NonLinearStaticControl;

   read_next_line();

   while ( !is_new_key() )
   {
      std::string word;
      get_field_lower_case( 1, word );

      if ( !interpret_incremental_iterative( word, *NLScontrol ) ) error_in_line++;

      read_next_line();
   }

   *control = NLScontrol;
   return 1;
}

int  input_data::read_nonlinear_dynamic( Control<MyDomain> **control )
{
   double alpha=0.25, delta=0.5;
   int    integration_type=1;

   NonLinearNewmarkControl *NMcontrol(0);
   ExplicitControl *EXcontrol(0);

   read_next_line();

   std::string word;
   get_field_lower_case( 1, word );

   if      ( word == "newmark" )
   {
      if ( nr_of_fields > 1 )
      {
         get_field( 2, &alpha );
         get_field( 3, &delta );
      }
      NMcontrol = new NonLinearNewmarkControl;
      NMcontrol->set_Newmark( alpha, delta );
   }
   else if ( word == "explicit" )
   {
      integration_type = 2;
      EXcontrol = new ExplicitControl;
   }
   else
   {
      NMcontrol = new NonLinearNewmarkControl;
      NMcontrol->set_Newmark( alpha, delta );
   }

   read_next_line();

   while ( !is_new_key() )
   {
      get_field_lower_case( 1, word );

      if      ( word == "rayleigh" )
      {
         double alpha, beta;
         get_field( 2, &alpha );
         get_field( 3, &beta );

         if ( integration_type == 1 )
            NMcontrol->set_Rayleigh_damping( alpha, beta );
         else if ( integration_type == 2 )
            EXcontrol->set_Rayleigh_damping( alpha, beta );
         else
         {
            cerr << "program error in read_nonlinear_dynamic" << endl;
            throw syntax_error();
         }
      }
      else
      {
         if ( integration_type == 1 )
         {
            if ( !interpret_incremental_iterative( word, *NMcontrol ) ) error_in_line++;
         }
         else if ( integration_type == 2 )
         {
            if ( !interpret_incremental( word, *EXcontrol ) ) error_in_line++;
         }
         else assert( false );
      }

      read_next_line();
   }

   if ( integration_type == 1 )
      *control = NMcontrol;
   else if ( integration_type == 2 )
      *control = EXcontrol;
   else
   {
      cerr << "program error in read_nonlinear_dynamic" << endl;
      throw syntax_error();
   }
   return 1;
}

bool input_data::interpret_incremental( const std::string &word, MyControl &control )
{ /* return true if the line was interpreted by this routine */
   double   timestep, size;
   int      nr_increments, load_set, max_resize;

   if (      word == "increments" )
   {
      get_field( 2, &nr_increments );
      control.set_increments( 1, 1.0/nr_increments, nr_increments );
   }
   else if ( word == "increment" )
   {
      if ( nr_of_fields != 7 )
      {
         error_in_line++;
         cerr << "Wrong increment load line" << endl <<
                 "Should be: increment load i size d steps n" << endl;
         return true;
      }
      std::string word2, word4, word6;
      get_field_lower_case( 2, word2 );
      get_field_lower_case( 4, word4 );
      get_field_lower_case( 6, word6 );

      if ( word2 == "load" && word4 == "size" && word6 == "steps" )
      {
         get_field( 3, &load_set );
         get_field( 5, &size );
         get_field( 7, &nr_increments );
         control.set_increments( load_set, size, nr_increments );
      }
      else
      {
         error_in_line++;
         cerr << "Wrong loading type" << endl;
      }

   }
   else if ( word == "timestep" )
   {
      get_field( 2, &timestep );
      control.set_time_increment( timestep );
   }
   else if ( word == "maxresize" )
   {
      get_field( 2, &max_resize );
      control.set_max_resize( max_resize );
   }
   else if ( word == "save" )
   {
      FeaTure::StepSelect selection;
      std::string word2;
      get_field_lower_case( 2, word2 );

      /* read step nr */
      if ( word2 == "step" )
      {
         int step;
         for ( int i=3; i<=nr_of_fields; i++ )
         {
            get_field( i, &step );
            selection.add_step( step );
         }
      }
      else if ( word2 == "all" )
         selection.add_all_steps();
      else if ( word2 == "last" )
         selection.add_last_step();

      control.set_restart_select( selection );
   }
   else if ( word == "restart" )
   {
      std::string word2;
      get_field_lower_case( 2, word2 );

      control.set_restart( word2 );
   }
   else return false;

   return true;
}

bool input_data::interpret_incremental_iterative( const std::string &word,
                  FeaTure::IncrementalIterativeControl<MyDomain> &control )
{
   double   force_norm;
   int      maxiter;

   if ( interpret_incremental( word, control ) ) return true;

   if      ( word == "full" )
      control.set_iteration_method( FULL_NEWTON_RAPHSON );
   else if ( word == "modified" )
      control.set_iteration_method( MODIFIED_NEWTON_RAPHSON );
   else if ( word == "unbalance" )
   {
      get_field( 2, &force_norm );
      control.set_unbalance_crit( force_norm );
   }
   else if ( word == "maxiter" )
   {
      get_field( 2, &maxiter );
      control.set_max_iter( maxiter );
   }
   else return false;

   return true;
}

int input_data::read_load_definition( MyDomain *model )
{
   int       set_nr;
   bool      next_set;
   LoadSet  *load_set;

   read_next_line();

   while ( !is_new_key() )
   {
      std::string word;
      get_field_lower_case( 1, word );

      /* read SET nr */
      if ( word != "set" || nr_of_fields != 2 )
      {
         cerr << "first line of each set in *LOADEF must define the SET number" << endl;
         error_in_line++;
         return 0;
      }
      get_field( 2, &set_nr );

      load_set = new LoadSet(set_nr);

      /* read loading type */

      next_set = false;
      read_next_line();

      do
      {
         std::string word;
         get_field_lower_case( 1, word );

         if ( word == "pressure" )
            read_pressure_load( model, load_set );
         else if ( word == "loads" )
            read_nodal_loads( model, load_set );
         else if ( word == "prescribed" )
            read_prescribed( model, load_set );
         else if ( word == "set" )
            next_set = true;
         else
         {
            cerr << "undefined load type in definition (*LOADEF)" << endl;
            error_in_line++;
            return 0;
         }
      } while ( !is_new_key() && !next_set );

      model->add_load_set( load_set );
   }

   return 1;
}

int input_data::read_post( Control<MyDomain> *control )
{
   FeaTure::PostProc  post;

   if ( control == NULL )
   {
      cerr << "*POST must be after *NONLINEAR" << endl;
      error_in_line++;
      read_next_line();
      return 0;
   }

   read_next_line();

   while ( !is_new_key() )
   {
      std::string word;
      get_field_lower_case( 1, word );

      /* read step nr */
      if ( word == "step" )
      {
         int step;
         for ( int i=2; i<=nr_of_fields; i++ )
         {
            get_field( i, &step );
            post.add_step( step );
         }
      }
      else if ( word == "all" )
         post.add_all_steps();
      else if ( word == "last" )
         post.add_last_step();
      else if ( word == "displacement" )
         post.set_displa();
      else if ( word == "temperature" )
         post.set_temp();
      else if ( word == "stress" )
         post.set_stress();
      else if ( word == "equivalent" )
         post.set_epspeq();
      else
      {
         cerr << "first word of each line in POST selection must be: step, all or last" << endl;
         error_in_line++;
         return 0;
      }

      read_next_line();
   }

   control->set_postproc( post );
   return 1;
}

int input_data::read_pressure_load( MyDomain *model, LoadSet *load_set )
{
   int       field, elm_nr, face_nr;
   FiniteElement* elmptr;
   double    magnitude;
   FeaTure::Pressure_Loads *pres_load;

   pres_load = new FeaTure::Pressure_Loads;

   read_next_line();
   field = 1;
   set_field(field);
   while ( isdigit( current_word[0] ) )
   {
      get_field( 1, &elm_nr );
      elmptr = model->get_element_ptr( elm_nr );
      if ( elmptr == NULL )
      {
         cerr << "pressure on unknown element number " << elm_nr << endl;
         error_in_line++;
         delete pres_load;
         return 0;
      }
      else
      {
         get_field( 2, &face_nr );
         get_field( 3, &magnitude );
         pres_load->add_load( elmptr, face_nr, magnitude );
      }
      read_next_line();
   }
   if ( pres_load != NULL ) load_set->add_load( pres_load );

   return 1;
}

void input_data::get_direction( const std::string &input, FeaTure::Direction *dir )
{
   int ret_val, dir_nr;

   if ( input.size() == 0 )
   {
      cerr << "direction missing\n";
      error_in_line++;
      return;
   }
   switch ( input[0] )
   {
      case 'x':
      case 'X': dir->set_dir( 1.0, 0.0, 0.0 ); break;
      case 'y':
      case 'Y': dir->set_dir( 0.0, 1.0, 0.0 ); break;
      case 'z':
      case 'Z': dir->set_dir( 0.0, 0.0, 1.0 ); break;
      default : ret_val = sscanf( input.c_str(), "%d", &dir_nr );
                if ( ret_val==1 )
                {
                   cerr << "free direction not implemented yet\n";
                   error_in_line++;
                }
                else
                {
                   error_in_line++;
                   cerr << "error in direction\n";
                }
   }
   return;
}

bool input_data::is_dof_type( char test, Dof::Type *type )
{
   test = (char)tolower( test );

   switch ( test )
   {
      case 'u':   *type = Dof::u; break;
      case 'r':   *type = Dof::r; break;
      case 't':   *type = Dof::t; break;
      default :   return false;
   }
   return true;
}

bool input_data::is_load_type( char test, Dof::Load *type )
{
   test = (char)tolower( test );

   switch ( test )
   {
      case 'f':   *type = Dof::f; break;
      case 'm':   *type = Dof::m; break;
      case 'q':   *type = Dof::q; break;
      default :   return false;
   }
   return true;
}

bool input_data::separator( char test )
{
   char t2[2];
   t2[0] = test;
   t2[1] = '\0';
   return ( strpbrk( t2, separators.c_str() ) != 0 );
}

char input_data::skip_blanks()
{
   char in_char = '\0';

   while ( !feof(in_ptr) && ( (in_char=(char)getc(in_ptr)) == ' ' || in_char=='\t' ) )
      in_char = ' ';

   if ( in_char == ' ' || in_char == '\t' ) return '\0';
   else return in_char;
}

void input_data::skip_line()
{
   while ( !feof(in_ptr) && ( getc(in_ptr) != '\n' ) );
}

char input_data::read_word( std::string *inputline )
{
   char  in_char('\0');

   while ( !feof(in_ptr) && !separator( in_char=(char)fgetc(in_ptr) ) )
      *inputline += in_char;

   if ( !separator( in_char ) ) in_char = '\0';

   return in_char;
}

void input_data::read_line()
/* get nr. of input fields and compressed input line */
{
   char in_char;

   nr_of_fields = 0;
   input_line.resize(0);

   in_char = skip_blanks();
   while ( in_char != '\n' && in_char != '\0' )
   {
      if ( in_char == CONT_CHAR )
      {
         skip_line();
         line_nr++;
      }
      else
      {
         if ( nr_of_fields != 0 ) input_line += ' ';
         input_line += in_char;

         in_char = read_word( &input_line );

         nr_of_fields++;
      }

      if ( in_char != '\0' && in_char != '\n' )
         in_char = skip_blanks();
   }
   line_nr++;

   return;
}

void input_data::print_key()
{
   switch ( key )
   {
      case TITLE    : cout << "read_title";       break;
      case ELEMENTS : cout << "read_elements";    break;
      case NODES    : cout << "read_nodes";       break;
      case SUPPRESS : cout << "read_suppress";    break;
      case MATERIAL : cout << "read_material";    break;
      case GEOMETRY : cout << "read_geometry";    break;
      case DATA     : cout << "read_data";        break;
      case NONLINEAR: cout << "read_nonlinear";   break;
      case LINEAR   : cout << "read_linear";      break;
      case END      : cout << "read_end";         break;
      case NOKEY    : cout << "no_key";           break;
   }
}

int input_data::nr_of_errors()
{
   return error_in_file;
}

int input_data::is_key()
{
   if ( input_line[0] != '*' ) return NOKEY;

   int   keyword;

   std::string word;
   get_field_lower_case( 1, word );

   if (      word == "*title"     ) keyword = TITLE;
   else if ( word == "*elements"  ) keyword = ELEMENTS;
   else if ( word == "*nodes"     ) keyword = NODES;
   else if ( word == "*suppress"  ) keyword = SUPPRESS;
   else if ( word == "*material"  ) keyword = MATERIAL;
   else if ( word == "*geometry"  ) keyword = GEOMETRY;
   else if ( word == "*nonlinear" ) keyword = NONLINEAR;
   else if ( word == "*linear"    ) keyword = LINEAR;
   else if ( word == "*eigen"     ) keyword = EIGEN;
   else if ( word == "*loadef"    ) keyword = LOADEF;
   else if ( word == "*post"      ) keyword = POST;
   else if ( word == "*end"       ) keyword = END;
   else
   {
      cerr << "unknown keyword: *" << word << endl;
      error_in_line++;
      keyword = NOKEY;
   }

   return keyword;
}

int input_data::is_element_name()
{
   int   keyword;

   std::string word;
   get_field_lower_case( 1, word );

   if (      word == "truss"      ) keyword = TRUSS;
   else if ( word == "beam2d"     ) keyword = BEAM2D;
   else if ( word == "tri3psn"    ) keyword = TRI3PSN;
   else if ( word == "tri3pss"    ) keyword = TRI3PSS;
   else if ( word == "tri3axi"    ) keyword = TRI3AXI;
   else if ( word == "tri3therm"  ) keyword = TRI3THERM;
   else if ( word == "tri3psntherm" ) keyword = TRI3PSNTHERM;
   else if ( word == "tri6psn"    ) keyword = TRI6PSN;
   else if ( word == "tri6pss"    ) keyword = TRI6PSS;
   else if ( word == "tri6axi"    ) keyword = TRI6AXI;
   else if ( word == "tri6psna"   ) keyword = TRI6PSNA;
   else if ( word == "tri6pssa"   ) keyword = TRI6PSSA;
   else if ( word == "tri6axia"   ) keyword = TRI6AXIA;
   else if ( word == "quad4psn"   ) keyword = QUAD4PSN;
   else if ( word == "quad4pss"   ) keyword = QUAD4PSS;
   else if ( word == "quad4axi"   ) keyword = QUAD4AXI;
   else if ( word == "quad4psstherm" ) keyword = QUAD4PSSTHERM;
   else if ( word == "quad4psntherm" ) keyword = QUAD4PSNTHERM;
   else if ( word == "tlquad4psn" ) keyword = TLQUAD4PSN;
   else if ( word == "quad8psn"   ) keyword = QUAD8PSN;
   else if ( word == "quad8pss"   ) keyword = QUAD8PSS;
   else if ( word == "quad8axi"   ) keyword = QUAD8AXI;
   else if ( word == "quad8psna"  ) keyword = QUAD8PSNA;
   else if ( word == "quad8pssa"  ) keyword = QUAD8PSSA;
   else if ( word == "quad8axia"  ) keyword = QUAD8AXIA;
   else if ( word == "tlquad8psn" ) keyword = TLQUAD8PSN;
   else if ( word == "hex8"       ) keyword = HEX8;
   else if ( word == "mass2d"     ) keyword = MASS2D;
   else if ( word == "mass3d"     ) keyword = MASS3D;
   else                             keyword = NOELM;

   return keyword;
}
