// Copyright (c) 1999, A.H. van den Boogaard

#include "thermal.h"

using FeaTure::Vector;
using FeaTure::Matrix;
using FeaTure::CollectArray;
using FeaTure::StressType;

LinearThermal_Point::LinearThermal_Point( LinearThermal_type *mat, StressType str )
                    :FeaTure::Thermal_Point( str ), parameters(mat)
{
   int dim = get_nodal_dimension( dimension_type );
   phi.resize(dim);
   newphi.resize(dim);
   phi.clear();
   newphi.clear();
   Temp = 0.0;
   newTemp = 0.0;
//   accumulation = 0.0;
}

LinearThermal_Point::~LinearThermal_Point(){}

Matrix LinearThermal_Point::get_conduction_matrix() const
{
   int dim = get_nodal_dimension( dimension_type );
   Matrix D(dim,dim,0);
   double lambda = parameters->get_conductivity();
   for ( int i=1; i<=dim; i++ ) D(i,i) = lambda;
   return D;
}

Vector LinearThermal_Point::get_flux() const
{
   return get_conduction_matrix()*newphi;
}

void LinearThermal_Point::set_flux_increment( const Vector& delta_phi )
{
   newphi = phi + delta_phi;
}

void LinearThermal_Point::set_potential_increment( double Tinc )
{
   newTemp = Temp + Tinc;
//   double c, rho, dT;
//   rho = parameters->get_density();
//   c = parameters->get_capacity();
////   dT = newTemp - Temp;
//   dT = Tinc;
//   accumulation = rho*c*dT;
   return;
}

void LinearThermal_Point::finish_increment()
{
   phi = newphi;
   Temp = newTemp;
   return;
}

double LinearThermal_Point::get_rho_c() const
{
   double c, rho, rho_c;
   rho = parameters->get_density();
   c = parameters->get_capacity();
   rho_c = rho*c;
   return rho_c;
}

double LinearThermal_Point::get_accumulation() const
{
   double c, rho, dT;
   rho = parameters->get_density();
   c = parameters->get_capacity();
   dT = newTemp - Temp;
   return rho*c*dT;
//   return accumulation;
}

void LinearThermal_Point::GetRestartData( CollectArray &data ) const
{
   data.reserve(phi.size()+1);
//   data.add(Temp);
   data.add(phi);
}

void LinearThermal_Point::PutRestartData( const CollectArray &data )
{
   Temp = data.extract();
   data.extract(phi);
   newTemp = Temp;
   newphi = phi;
}

LinearThermal_type::LinearThermal_type( int matnr ):FeaTure::Material_type()
{
   nr = matnr;
   lambda = 0;
   capacity = 0;
}

LinearThermal_type::~LinearThermal_type(){}

int LinearThermal_type::set_conductivity( double lam )
{
   lambda = lam;
   if ( lambda <= 0.0 )
   {
      std::cerr << "conduction <= 0; this is not allowed" << std::endl;
      return 1;
   }
   return 0;
}

double LinearThermal_type::get_conductivity() const { return lambda; }

int LinearThermal_type::set_capacity( double c )
{
   capacity = c;
   if ( capacity <= 0.0 )
   {
      std::cerr << "capacity <= 0; this is not allowed" << std::endl;
      return 1;
   }
   return 0;
}

double LinearThermal_type::get_capacity() const { return capacity; }

#ifdef NOVARIABLERETURN
FeaTure::Material_Point* LinearThermal_type::new_Material_Point( StressType st )
#else
LinearThermal_Point* LinearThermal_type::new_Material_Point( StressType st )
#endif
{
   return new LinearThermal_Point(this,st);
}
