// Copyright (c) 1999,2003, A.H. van den Boogaard

#include "cartvect.h"
#include "ftexcept.h"

#include <iomanip>
#include <cmath>

using FeaTure::CartVector;
using FeaTure::Direction;
using FeaTure::Location;
using std::cout;
using std::setw;
using std::sqrt;

CartVector::CartVector( const CartVector& vec )
{
   x[0] = vec.x[0];
   x[1] = vec.x[1];
   x[2] = vec.x[2];
   return;
}

CartVector::CartVector( double x1, double x2, double x3 )
{
   x[0]=x1; x[1]=x2; x[2]=x3;
   return;
}

void CartVector::print() const
{
   cout << setw(12) << x[0] << " " << setw(12) << x[1]
                            << " " << setw(12) << x[2];
}

std::ostream& FeaTure::operator<<( std::ostream& os, const CartVector& vec )
{
   double xx, yy, zz;
   int w = os.width(0);

   vec.getValues( xx, yy, zz );

   return os << setw(w) << xx << ' '
             << setw(w) << yy << ' '
             << setw(w) << zz ; 
}

CartVector CartVector::add(const CartVector& src2) const {
  int i;

  CartVector m;

  for (i=0; i<3; i++) m.x[i] = x[i] + src2.x[i];
 
  return m;
}

CartVector CartVector::selfadd(const CartVector& src2) {

  for (int i=0; i<3; i++) x[i] += src2.x[i];  
  return *this;
}

CartVector CartVector::sub(const CartVector& src2) const {
  int i;

  CartVector m;

  for (i=0; i<3; i++) m.x[i] = x[i] - src2.x[i];

  return m;
}

CartVector CartVector::selfsub(const CartVector& src2) {

  for (int i=0; i<3; i++) x[i] -= src2.x[i];  
  return *this;
}

CartVector CartVector::mul(double val) const {
  CartVector cv(*this);
  for (int i=0; i<3; i++) cv.x[i] *= val;
  return cv;
}

CartVector CartVector::selfmult(double scalar) {
  x[0] *= scalar; x[1] *= scalar; x[2] *= scalar;
  return *this;
}

void CartVector::getValues( double &xx, double &yy, double &zz ) const
{
   xx = x[0];
   yy = x[1];
   zz = x[2];
}

double CartVector::dot(const CartVector& vec2) const {
  return ( x[0]*vec2.x[0]+x[1]*vec2.x[1]+x[2]*vec2.x[2] );
}

double CartVector::normsq() const {
  return ( x[0]*x[0]+x[1]*x[1]+x[2]*x[2] );
}

double CartVector::norm() const {
  return ( sqrt( x[0]*x[0]+x[1]*x[1]+x[2]*x[2] ) );
}

Direction::Direction( const CartVector& vect ):CartVector( vect )
{
   double length = norm();
   if ( length > 0.0 )
      selfmult( 1/length );
   else
      throw program_error( "illegal direction, length = 0" );
}

void Direction::set_dir( double xx, double yy, double zz )
{
   double lengthsq, factor;

   lengthsq = xx*xx + yy*yy + zz*zz;
   if ( lengthsq != 1.0 )
   {
      if ( lengthsq > 0.0 )
      {
         factor = 1.0 / sqrt( lengthsq );
         xx *= factor;
         yy *= factor;
         zz *= factor;
      }
      else
         throw program_error( "illegal direction, length = 0" );
   }

   x[0] = xx;
   x[1] = yy;
   x[2] = zz;

}

void Direction::get_dir( double& xx, double& yy, double& zz ) const
{
   xx = x[0];
   yy = x[1];
   zz = x[2];
}

void Direction::print() const
{
   cout.setf( std::ios::fixed, std::ios::floatfield );
   cout.precision(4);
   cout << setw(7) << x[0] << " " << setw(7) << x[1]
                           << " " << setw(7) << x[2];
}

Location::Location( double xx, double yy, double zz )
{
   x[0] = xx;
   x[1] = yy;
   x[2] = zz;
}

void Location::set_location( double xx, double yy, double zz )
{
   x[0] = xx;
   x[1] = yy;
   x[2] = zz;
}

void Location::get_location( double& xx, double& yy, double& zz) const
{
   xx = x[0];
   yy = x[1];
   zz = x[2];
}

void Location::print() const
{
   cout.setf( std::ios::scientific, std::ios::floatfield );
   cout.precision(4);
   cout << setw(12) << x[0] << " " << setw(12) << x[1]
                            << " " << setw(12) << x[2];
}

CartVector FeaTure::outer( const CartVector& a, const CartVector& b )
{
   double c0, c1, c2;

   c0 =  a.x[1]*b.x[2]-a.x[2]*b.x[1];
   c1 = -a.x[0]*b.x[2]+a.x[2]*b.x[0];
   c2 =  a.x[0]*b.x[1]-a.x[1]*b.x[0];

   CartVector c(c0,c1,c2);
   return c;
}
