// Copyright (c) 1999, A.H. van den Boogaard

#include "elembase.h"
#include "ftexcept.h"
#include <sstream>

using FeaTure::Vector;
using FeaTure::Matrix;
using FeaTure::SystemVector;
using FeaTure::SystemMat;
using FeaTure::ElementBase;
using FeaTure::PostProc;

// initialize static data
double ElementBase::time_increment = 0;
PostProc::PostItem ElementBase::post_item = PostProc::stresses;

ElementBase::ElementBase( int nr, Node_ptr *ElNodes, int nr_of_ElNodes, int nr_dofs )
                            :nr_of_element_nodes(nr_of_ElNodes),elm_nr(nr),Nodes(ElNodes)
{
   nr_of_local_dofs = nr_dofs;
   dofs             = NULL;
   nr_of_system_dofs = 0;
}

ElementBase::ElementBase()
{
   elm_nr = nr_of_element_nodes = nr_of_local_dofs = 0;
   Nodes = NULL;
   dofs = NULL;
   nr_of_system_dofs = 0;
}

ElementBase::~ElementBase()
{
   delete [] Nodes;
   delete [] dofs;
}

Vector ElementBase::GetLocalPresVec( int /*surface*/, double /*magnitude*/ ) const
{
   std::ostringstream message; 
   message << "Pressure load on element " << get_nr() << " is not implemented";
   throw not_implemented_error( message.str() );

   Vector dummy;
   return dummy;
}

void ElementBase::add_dofs_to_nodes() const
{
   LocalDof* ldofs = GetLocalDofs();

   for ( int idof=0; idof<nr_of_local_dofs; idof++ )
      Nodes[ldofs[idof].loc_node_nr-1]->create_dof( Dof::free_dof,
                                       ldofs[idof].phys_type, ldofs[idof].dir );

   delete [] ldofs;
}

void ElementBase::CreateGlobalDofs()
{
   dof_sets  system_dofs;
   int       *syst_upto_local, idof;

   syst_upto_local = new int[nr_of_local_dofs];

   LocalDof* ldofs = GetLocalDofs();

   for ( idof=0; idof<nr_of_local_dofs; idof++ )
   {
      Nodes[ldofs[idof].loc_node_nr-1]->get_dof( ldofs[idof].phys_type,
                                                ldofs[idof].dir, &system_dofs );
      syst_upto_local[idof] = system_dofs.nr_of_dofs();
   }

   nr_of_system_dofs = system_dofs.nr_of_dofs();
   dofs = new Dof_ptr[nr_of_system_dofs];
   for ( idof=0; idof<nr_of_system_dofs; idof++ )
      dofs[idof] = system_dofs.dof(idof);

   transmat.setSparse( nr_of_local_dofs, nr_of_system_dofs, nr_of_system_dofs );
   int strtcol=0;
   for ( int row=0; row<nr_of_local_dofs; row++ )
   {
      int stopcol=syst_upto_local[row];
      for ( int col=strtcol; col<stopcol; col++ )
         transmat.put( row+1,col+1,system_dofs.contribution(col) );
      strtcol=stopcol;
   }

   delete [] ldofs;

   return;
}

Vector ElementBase::GetGlobalLHS() const
{
   return transmat.tmult( GetLocalLHS() );
}

void ElementBase::InitStiff_in_Global( SystemMat *syst ) const
{
   for ( int row=0; row<nr_of_system_dofs; row++ )
   {
      for ( int col=0; col<nr_of_system_dofs; col++ )
      {
         if ( dofs[row]->is_free() && dofs[col]->is_free() )
            syst->init(dofs[row]->get_global_dof(), dofs[col]->get_global_dof() );
      }
   }
}

void ElementBase::PutStiff_and_RHS_in_Global( SystemMat *syst, Vector& rhs ) const
{
   bool expand = NeedExpandStrain();
   Vector local_stress_correction;
   Vector global_stress_correction;

   Matrix elm_glob_mat = transmat.preTmult_postMult( GetLocalStiff( local_stress_correction ) );
   if ( expand ) global_stress_correction = transmat.tmult( local_stress_correction );

   for ( int row=0; row<nr_of_system_dofs; row++ )
   {
      for ( int col=0; col<nr_of_system_dofs; col++ )
      {
         if ( dofs[row]->is_free() )
         {
            if ( dofs[col]->is_free() )
               syst->add(dofs[row]->get_global_dof(), dofs[col]->get_global_dof(),
                         elm_glob_mat[row][col] );
            else if ( dofs[col]->is_prescribed() )
               rhs[ dofs[row]->get_global_dof()-1 ] -=
                  elm_glob_mat[row][col]*(dofs[col]->get_increment());
         }
      }
      if ( expand && dofs[row]->is_free() )
         rhs[dofs[row]->get_global_dof()-1] += global_stress_correction[row];
   }
}

void ElementBase::PutStiff_in_Global( SystemMat *syst, Vector& rhs ) const
{
   bool expand = NeedExpandStrain();
   Vector local_stress_correction;
   Vector global_stress_correction;

   Matrix elm_glob_mat = transmat.preTmult_postMult( GetLocalStiff( local_stress_correction ) );
   if ( expand ) global_stress_correction = transmat.tmult( local_stress_correction );

   for ( int row=0; row<nr_of_system_dofs; row++ )
   {
      for ( int col=0; col<nr_of_system_dofs; col++ )
      {
         if ( dofs[row]->is_free() )
         {
            if ( dofs[col]->is_free() )
               syst->add(dofs[row]->get_global_dof(), dofs[col]->get_global_dof(),
                         elm_glob_mat[row][col] );
         }
      }
      if ( expand && dofs[row]->is_free() )
         rhs[dofs[row]->get_global_dof()-1] += global_stress_correction[row];
   }
}

Vector ElementBase::get_local_total_vector() const
{
   Vector U_global(nr_of_system_dofs);

   for ( int idof=0; idof<nr_of_system_dofs; idof++ )
      U_global[idof] = dofs[idof]->get_total();

   return transmat*U_global;
}

Vector ElementBase::get_local_increment_vector() const
{
   Vector U_global(nr_of_system_dofs);

   for ( int idof=0; idof<nr_of_system_dofs; idof++ )
      U_global[idof] = dofs[idof]->get_increment();

   return transmat*U_global;
}

void ElementBase::PutLHS_in_Global( SystemVector& lhs ) const
{
   Vector LHS(GetGlobalLHS());
   for ( int i=0; i<nr_of_system_dofs; i++ )
      lhs( *(dofs[i]) ) += LHS[i];
}

Matrix ElementBase::GetLocalMass( bool /*lumped*/ ) const
{
   Matrix M(nr_of_local_dofs,nr_of_local_dofs,0.0);
   return M;
}

void ElementBase::InitMass_in_Global( SystemMat *syst, bool lumped ) const
{
   if ( lumped )
   {
      for ( int row=0; row<nr_of_system_dofs; row++ )
         if ( dofs[row]->is_free() )
               syst->init(dofs[row]->get_global_dof(), dofs[row]->get_global_dof() );
   }
   else
   {
      for ( int row=0; row<nr_of_system_dofs; row++ )
         for ( int col=0; col<nr_of_system_dofs; col++ )
            if ( dofs[row]->is_free() && dofs[col]->is_free() )
               syst->init(dofs[row]->get_global_dof(), dofs[col]->get_global_dof() );
   }
}

void ElementBase::PutMass_in_Global( SystemMat *syst, bool lumped ) const
{
   Matrix elm_M_glob = transmat.preTmult_postMult( GetLocalMass( lumped ) );

   if ( lumped )
   {
      for ( int row=0; row<nr_of_system_dofs; row++ )
         if ( dofs[row]->is_free() )
            syst->add(dofs[row]->get_global_dof(), dofs[row]->get_global_dof(),
                      elm_M_glob[row][row] );
   }
   else
   {
      for ( int row=0; row<nr_of_system_dofs; row++ )
         for ( int col=0; col<nr_of_system_dofs; col++ )
            if ( dofs[row]->is_free() && dofs[col]->is_free() )
               syst->add(dofs[row]->get_global_dof(), dofs[col]->get_global_dof(),
                         elm_M_glob[row][col] );
   }
}

void ElementBase::AddPres2Load( int surface, double magnitude,
                    SystemVector& load ) const
{
   Vector ElmPresLoad( transmat.tmult( GetLocalPresVec( surface, magnitude ) ) );
   for ( int i=0; i<nr_of_system_dofs; i++ )
   {
      load( (*dofs[i]) ) += ElmPresLoad[i];
   }
}

void ElementBase::WriteRestartData( std::ostream &restart ) const
{
   restart << 'E' << get_nr() << std::endl;
   WriteElementRestartData( restart );
}

void ElementBase::ReadRestartData( std::istream &restart )
{
   char c;
   int  n;

   restart >> c >> n;
   if ( c != 'E' || get_nr() != n ) throw file_error( "corrupt restart file" );
   ReadElementRestartData( restart );
}
