// Copyright (c) 2003, A.H. van den Boogaard

#include "ftmatrix.h"
#include "ftexcept.h"
#include <iomanip>
#include <sstream>

using FeaTure::Matrix;
using FeaTure::Vector;

Matrix Matrix::without_rowcol( size_t rowcol ) const
{
   // Create a matrix from the current matrix, without row
   // and column rowcol

   if ( rowcol < 1 || rowcol > m_ || rowcol > n_ )
   {
      std::ostringstream message;
      message << "can not leave out row and column " << rowcol <<
                 " of matrix of size " << m_ <<"*"<< n_;
      throw dimension_error( message.str() );
   }

   Matrix m(m_-1,n_-1);
   double *to_ptr( m.v_ );
   double *from_ptr( v_ );

   for ( size_t i=0; i<m_; i++ )
      if ( i == rowcol-1 )
         from_ptr += n_;
      else
         for ( size_t j=0; j<n_; j++ )
            if ( j == rowcol-1 )
               from_ptr++;
            else
               *to_ptr++ = *from_ptr++;

   return m;
}

Matrix Matrix::add_SubMat( size_t subrow, size_t subcol,
                           const Matrix& SubMat )
{
  // Add SubMat to this matrix, where SubMat(1,1) coincides with
  // Matrix(subrow,subcol)

  if (n_ < SubMat.n_+subcol-1 || m_ < SubMat.m_+subrow-1   )
  {
     std::ostringstream message;
     message << "add_SubMat: Incompatible sizes: left=[" << m_ << "]["
             << n_ << "] right=[" << ( SubMat.m_+subrow-1 )
             << "][" << ( SubMat.n_+subcol-1 ) << "].";
     throw dimension_error( message.str() );
  }

  // transform to 0-based
  subcol--;
  subrow--;
  for ( size_t i=0; i<SubMat.m_; i++ )
     for ( size_t j=0; j<SubMat.n_; j++ )
        v_[(subrow+i)*n_+subcol+j] += SubMat.v_[i*SubMat.n_+j];

  return *this;
}

Matrix Matrix::inv() const
{
   // return the inverse of this matrix

   size_t size( m_ );

   if (size != n_) {
      std::ostringstream message;
      message << "inv: error, trying to invert [" << m_ << "][" << n_
              << "] matrix.";
      throw dimension_error( message.str() );
   }

   if ( size<4 )
   {
      // for efficiency write out
      Matrix M_inv( size, size );
      double det;

      switch ( size )
      {
         case 0: break;
         case 1: det = v_[0];
                 if ( det==0 ) throw matrix_error(
                           "determinant zero in calculation of inverse matrix");
                 M_inv.v_[0] = 1.0 / det;
                 break;
         case 2: det = v_[0]*v_[3]-v_[1]*v_[2];
                 if ( det==0 ) throw matrix_error(
                           "determinant zero in calculation of inverse matrix");
                 M_inv.v_[0] =  v_[3]/det;
                 M_inv.v_[1] = -v_[1]/det;
                 M_inv.v_[2] = -v_[2]/det;
                 M_inv.v_[3] =  v_[0]/det;
                 break;
         case 3: {
                 double Sub11, Sub12, Sub13;
                 Sub11 = v_[4]*v_[8] - v_[5]*v_[7];
                 Sub12 = v_[3]*v_[8] - v_[6]*v_[5];
                 Sub13 = v_[3]*v_[7] - v_[6]*v_[4];
                 det = v_[0]*Sub11 - v_[1]*Sub12 + v_[2]*Sub13;
                 if ( det==0 ) throw matrix_error(
                           "determinant zero in calculation of inverse matrix");
                 M_inv.v_[0] =  Sub11/det;
                 M_inv.v_[1] = -(v_[1]*v_[8]-v_[2]*v_[7])/det;
                 M_inv.v_[2] =  (v_[1]*v_[5]-v_[2]*v_[4])/det;
                 M_inv.v_[3] = -Sub12/det;
                 M_inv.v_[4] =  (v_[0]*v_[8]-v_[2]*v_[6])/det;
                 M_inv.v_[5] = -(v_[0]*v_[5]-v_[2]*v_[3])/det;
                 M_inv.v_[6] =  Sub13/det;
                 M_inv.v_[7] = -(v_[0]*v_[7]-v_[1]*v_[6])/det;
                 M_inv.v_[8] =  (v_[0]*v_[4]-v_[1]*v_[3])/det;
                 }
                 break;
         default: throw dimension_error( "internal error in Matrix::inv()" );
      }

      return M_inv;
   }

  // General inversion
  // See page 171 of 'Introduction to numerical analysis'
  // by J. Stoer & R. Bulirsch

  Matrix m(*this);
  size_t n(m_);
  size_t *p = new size_t[n];   // will contain the column permutations
  double tmp, hlp;
  double * hv = new double[n];
  size_t tmpint;

  for (size_t i=0; i<n; i++) p[i]=i;

  for (size_t j=0; j<n; j++)
  {
    double max(std::fabs(m[j][j]));
    size_t r(j);
    for (size_t i=j+1; i<n; i++) {
      if ((std::fabs(tmp=m[i][j])) > max) {
        max=tmp;
        r=i;
      }
    }

    if (max==0)
       throw matrix_error( "determinant zero in calculation of inverse matrix");

    if (r>j) {
      for (size_t k=0; k<n; k++) {
        tmp = m[j][k];
        m[j][k] = m[r][k];
        m[r][k] = tmp;
        }
      tmpint = p[j];
      p[j] = p[r];
      p[r] = tmpint;
    }

    tmp = 1/m[j][j];
    for (size_t i=0; i<n; i++) {
      m[i][j] = tmp * m[i][j];
    }
    m[j][j] = tmp;

    for (size_t k=0; k<n; k++) {
      if (k==j) continue;
      for (size_t i=0; i<n; i++) {
        if (i==j) continue;
        hlp = m[i][k] - (m[i][j] * m[j][k]);
        m[i][k] = hlp;
      }
      m[j][k] = -tmp * m[j][k];
    }
  }

  for (size_t i=0; i<n; i++) {
    for (size_t k=0; k<n; k++) {
      hv[p[k]]=m[i][k];
    }
    for (size_t k=0; k<n; k++) {
      m[i][k] = hv[k];
    }
  }

  delete [] hv;
  delete [] p;
  return (m);
}

double Matrix::determinant() const
{
  assert (m_==n_);

  switch ( m_ ) {
    case 0:  return 0.0;
    case 1:  return v_[0];
    case 2:  return v_[0]*v_[3]-v_[1]*v_[2];
    case 3:  return v_[0]*(v_[4]*v_[8]-v_[5]*v_[7])
                   -v_[1]*(v_[3]*v_[8]-v_[5]*v_[6])
                   +v_[2]*(v_[3]*v_[7]-v_[4]*v_[6]);
    default: { std::ostringstream message;
           message << "determinant: error, trying to calculate determinant of ["
                   << m_ << "][" << n_ << "] matrix.";
               throw not_implemented_error( message.str() );
             }
  }
}

Vector Vector::add_SubVec( size_t subrow, const Vector& SubVec )
{
  if ( m_ < SubVec.m_+subrow-1 )
  {
     std::ostringstream message;
     message << "add_SubVec: Incompatible sizes: left=[" << m_ << "] right=["
             << ( SubVec.m_+subrow-1 ) << "].";
     throw dimension_error( message.str() );
  }

  // transform to 0-based
  subrow--;

  for ( size_t i=0; i<SubVec.m_; i++ )
        v_[subrow+i] += SubVec.v_[i];

  return *this;
}

void Vector::print_row(std::ostream &s) const
{
    int w = s.width();

    for (size_t i=0; i<m_; i++)
    {
        s  << std::setw(w) << v_[i];
        if ( i < m_-1 ) s  << " " ;
    }
}

std::ostream& FeaTure::operator<<(std::ostream &s, const Matrix &A)
{
   int w=s.width();
   size_t M(A.size1());
   size_t N(A.size2());
   for ( size_t i=0; i<M; i++ )
   {
      for ( size_t j=0; j<N; j++ )
         s << std::setw(w) << A[i][j];
      s << std::endl;
   }
   return s;
}

std::ostream& FeaTure::operator<<(std::ostream &s, const Vector &a)
{
   int w=s.width();
   size_t M(a.size());
   for ( size_t i=0; i<M; i++ )
      s << std::setw(w) << a[i] << std::endl;
   return s;
}

