// Copyright (c) 1999, A.H. van den Boogaard

#include "nonsym.h"
#include "ftexcept.h"

#include <iomanip>
#include <sstream>

using FeaTure::SkyNonSymMat;
using FeaTure::Vector;

using std::cout;
using std::endl;

SkyNonSymMat::SkyNonSymMat( size_t n ):SystemMat()
{
   size = n;
   p = new size_t[size];
   for ( size_t i=0; i<size; i++ ) p[i] = i;
   L = U = NULL;
   decomposed = false;
}

SkyNonSymMat::~SkyNonSymMat()
{
   if ( L != NULL )
   {
      for ( size_t i=0; i<size; i++ )
      {
         delete [] L[i];
         delete [] U[i];
      }
      delete [] L;
      delete [] U;
   }
   if ( p != NULL ) delete [] p;
}

void SkyNonSymMat::init(size_t row, size_t col)
{
   if ( L!=NULL )
      throw program_error( "matrix geometry can not be updated after creation" );

   if ( row > size || col > size || row < 1 || col < 1 )
   {
      std::ostringstream message;
      message << "matrix element (" << row << ',' << col
              << ") does not fit in matrix of size " << size;
      throw dimension_error( message.str() );
   }
   if ( col >= row )
   {  if ( row <= p[col-1] ) p[col-1] = row-1; }
   else
   {  if ( col <= p[row-1] ) p[row-1] = col-1; }

   return;
}

void SkyNonSymMat::clear()
{
   if ( L==NULL ) create_mem();

   for ( size_t i=0; i<size; i++ )
   {
      size_t maxj( i-p[i] );
      for ( size_t j=0; j<maxj; j++ )
      {
         L[i][j] = 0;
         U[i][j] = 0;
      }
      U[i][i-p[i]] = 0;
   }

   decomposed = false;
}

void SkyNonSymMat::create_mem()
{
   if ( size == 0 ) return;

   if ( L != NULL )
   {
      for ( size_t i=0; i<size; i++ )
      {
         delete [] L[i];
         delete [] U[i];
      }
      delete [] L;
      delete [] U;
   }

   L = new double*[size];
   U = new double*[size];
   for ( size_t i=0; i<size; i++ )
   {
      L[i] = new double[i-p[i]];
      U[i] = new double[i-p[i]+1];
   }
}

void SkyNonSymMat::add(size_t row, size_t col, double value)
{
   if ( L==NULL ) create_mem();

   if ( decomposed )
      throw program_error( "can not add matrix component to a decomposed matrix" );

   if ( row > size || col > size || row < 1 || col < 1 )
   {
      std::ostringstream message;
      message << "matrix element (" << row << ',' << col
              << ") does not fit in matrix of size " << size;
      throw dimension_error( message.str() );
   }
   if ( col >= row ) U[col-1][row-1-p[col-1]] += value;
   else              L[row-1][col-1-p[row-1]] += value;
}

void SkyNonSymMat::print( char *name ) const
{
   if ( decomposed ) cout << "Decomposed ";
   cout << name << endl << endl;

   cout.precision(3);
   cout.setf( std::ios::scientific, std::ios::floatfield );
   cout << "size =" << size << endl;
   if ( size > 10 ) cout << "(only left upper 10x10 part)" << endl;

   for ( size_t row = 0; row < size && row < 10 ; row++ )
   {
      for ( size_t col = 0; col < size && col < 10; col++ )
         cout << " " << std::setw(10) << dget(row,col);

      cout << endl;
   }
}

void SkyNonSymMat::solve( const Vector& rhs, Vector& x )
{
   if ( !decomposed )
   {
      if ( LU_factor() !=0 )
          throw matrix_error( "LU_factor() failed." );

      else decomposed = true;
   }

   x = rhs;
   if ( LU_solve( x ) != 0 )
       throw matrix_error( "LU_Solve() failed." );
}

double SkyNonSymMat::dget(size_t i, size_t j) const
{
   if ( i<=j ) // Upper triangular matrix
   {
      if ( i<p[j] ) return 0;
      else return U[j][i-p[j]];
   }
   else        // Lower triangular matrix
   {
      if ( j<p[i] ) return 0;
      else return L[i][j-p[i]];
   }
}

void SkyNonSymMat::dput(size_t i, size_t j, double a)
{
   if ( i<=j ) // Upper triangular matrix
   {
      assert( j>=p[i] );
      U[j][i-p[j]] = a;
   }
   else        // Lower triangular matrix
   {
      assert( i>=p[j] );
      L[i][j-p[i]] = a;
   }
}

int SkyNonSymMat::LU_factor()
{
/*
     Crout LU-factorization for a skyline storage non-symmetric
     matrix with symmetric non-zero structure, without pivoting.
*/
   for ( size_t j=1; j<size; j++ )
   {
      size_t pj(p[j]);
      for ( size_t i=pj; i<j; i++ )
      {
         double Uij, Lji;
         Uij = dget(i,j);
         Lji = dget(j,i);
         size_t pi(p[i]);
         double *Ljm, *Lim, *Umi, *Umj;
         size_t start = std::max(pi,pj);
         Ljm = &L[j][start-pj];
         Lim = &L[i][start-pi];
         Umi = &U[i][start-pi];
         Umj = &U[j][start-pj];

         for ( size_t m=start; m<i; m++ )
         {  // core of the analysis, optimized for speed
            Lji -= (*Ljm++)*(*Umi++); // dget(j,m)*dget(m,i);
            Uij -= (*Lim++)*(*Umj++); // dget(i,m)*dget(m,j);
         }
         double Uii = dget(i,i);
         if ( std::fabs(Uii) < 1.e-15 ) return 1;
         Lji /= Uii;

         dput(j,i,Lji);
         dput(i,j,Uij);
      }

      double Ujj(dget(j,j));
      for ( size_t m=p[j]; m<j; m++ ) Ujj -= dget(j,m)*dget(m,j);
      dput(j,j,Ujj);
   }
   return 0;
}

int SkyNonSymMat::LU_solve( Vector& x ) const
{
   if ( !decomposed ) return 1;
   if ( size == 0 ) return 0;

   for ( size_t i=1; i<size; i++ ) // forward substitution
   {
      for ( size_t j=p[i]; j<i; j++ ) x[i] -= dget(i,j)*x[j];
   }

   // warning: don't substitute int by size_t because of i>=0 check
   for ( size_t i=size-1; i>0; i-- ) // backward substitution (by column)
   {
      x[i] /= dget(i,i);
      for ( size_t j=p[i]; j<i; j++ ) x[j] -= dget(j,i)*x[i];
   }
   x[0] /= dget(0,0);

   return 0;
}

