#include "triangular.h"

using FeaTure::Matrix;
using FeaTure::Vector;

using std::sqrt;

void FeaTure::LDLtDecomposition( Matrix *Ap )
{
   //           T
   // Create LDL  decomposition of A
   // Algorithm 4.1.2 in Golub & VanLoan, Matrix Computations

   Matrix &A( *Ap );
   size_t n( A.size1() );
   assert( A.size2()==n );
   Vector v(n);

   for ( size_t j=0; j<n; j++ )
   {
      for ( size_t i=0; i<j; i++ )
         v[i] = A[j][i]*A[i][i];

      v[j] = A[j][j];
      for ( size_t i=0; i<j; i++ ) v[j] -= A[j][i]*v[i];
      A[j][j] = v[j];
      for ( size_t i=j+1; i<n; i++ )
      {
         double a(0);
         for ( size_t k=0; k<j; k++ ) a += A[i][k]*v[k];
         A[i][j] = (A[i][j]-a)/v[j];
      }
   }
}

void FeaTure::LDLtSubstitution( const Matrix &LDLt, Vector *xp )
{
   // forward and backward substitution to solve Ax=b
   // initially b is stored in x, on exit x is the solution x
   // LDLt is the decomposed matrix A

   Vector &x( *xp );
   size_t n( LDLt.size1() );
   assert( LDLt.size2()==n && x.size()==n );
   for ( size_t i=0; i<n; i++ )
      for ( size_t j=0; j<i; j++ ) x[i] -= LDLt[i][j]*x[j];

   for ( size_t i=0; i<n; i++ ) x[i]/=LDLt[i][i];

   for ( size_t i=n-1;; i-- )
   {
      for ( size_t j=i+1; j<n; j++ ) x[i] -= LDLt[j][i]*x[j];
      if ( i==0 ) break;
   }
}

void FeaTure::CholeskyDecomposition( Matrix *Ap )
{
   //          T
   // Create LL  (Cholesky) decomposition of A
   // Algorithm 4.2.2 in Golub & VanLoan, Matrix Computations

   Matrix &A( *Ap );
   size_t n( A.size1() );
   assert( A.size2()==n );

   for ( size_t k=0; k<n; k++ )
   {
      double Akk = sqrt( A[k][k] );
      A[k][k] = Akk;

      for ( size_t i=k+1; i<n; i++ )
         A[i][k]/=Akk;

      for ( size_t j=k+1; j<n; j++ )
      {
         double Ajk = A[j][k];
         for ( size_t i=j; i<n; i++ )
            A[i][j] -= A[i][k]*Ajk;
      }
   }
}

void FeaTure::CholeskySubstitution( const Matrix &L, Vector *xp )
{
   // forward and backward substitution to solve Ax=b
   // initially b is stored in x, on exit x is the solution x
   // L is the lower left Cholesky decomposed matrix A=LLt

   Vector &x( *xp );
   size_t n( L.size1() );
   assert( L.size2()==n && x.size()==n );
   for ( size_t i=0; i<n; i++ )
   {
      for ( size_t j=0; j<i; j++ ) x[i] -= L[i][j]*x[j];
      x[i] /= L[i][i];
   }

   for ( size_t i=n-1;; i-- )
   {
      for ( size_t j=i+1; j<n; j++ ) x[i] -= L[j][i]*x[j];
      x[i] /= L[i][i];
      if ( i==0 ) break;
   }
}


// Triangular matrix multiplications and inverse matrix multiplications

void FeaTure::PostMultLowerTInv( Matrix *Ap, const Matrix &L )
{
   // L is a lower left triangular matrix L(i,j) j>=i
   // the upper right part is not used and need not be initialized
   //                  -T
   // On exit  A <- A L

   Matrix &A( *Ap );
   size_t n( L.size1() );
   assert( L.size2()==n && A.size2()==n );
   size_t m( A.size1() );

   for ( size_t k=0; k<n; k++ )
   {
      for ( size_t i=0; i<m; i++ )
      {
         for ( size_t j=0; j<k; j++ ) A[i][k] -= A[i][j]*L[k][j];
         A[i][k] /= L[k][k];
      }
   }
}

void FeaTure::PreMultLowerInv( const Matrix &L, Matrix *Ap )
{
   // L is a lower left triangular matrix L(i,j) j>=i
   // the upper right part is not used and need not be initialized
   //               -1
   // On exit  A = L  A

   Matrix &A( *Ap );
   size_t n( L.size1() );
   assert( L.size2()==n && A.size1()==n );
   size_t m( A.size2() );

   for ( size_t k=0; k<m; k++ )
   {
      for ( size_t i=0; i<n; i++ )
      {
         for ( size_t j=0; j<i; j++ ) A[i][k] -= L[i][j]*A[j][k];
         A[i][k] /= L[i][i];
      }
   }
}

void FeaTure::PreMultLowerTInv( const Matrix &L, Matrix *Ap )
{
   // L is a lower left triangular matrix L(i,j) j>=i
   // the upper right part is not used and need not be initialized
   //               -T
   // On exit  A = L  A

   Matrix &A( *Ap );
   size_t n( L.size1() );
   assert( L.size2()==n && A.size1()==n );
   size_t m( A.size2() );

   for ( size_t k=0; k<m; k++ )
   {
      for ( size_t i=n-1;; i-- )
      {
         for ( size_t j=i+1; j<n; j++ ) A[i][k] -= L[j][i]*A[j][k];
         A[i][k] /= L[i][i];
         if ( i==0 ) break;
      }
   }
}

Matrix FeaTure::PreMultLower( const Matrix &L, const Matrix &A )
{
   // L is a lower left triangular matrix L(i,j) j>=i
   // the upper right part is not used and need not be initialized
   // Calculate R=L*A

   size_t n( L.size1() );
   assert( L.size2()==n && A.size1()==n );
   size_t m( A.size2() );

   Matrix R(n,m,0.0);

   for ( size_t i=0; i<n; i++ )
      for ( size_t j=0; j<m; j++ )
         for ( size_t k=0; k<=i; k++ ) R[i][j] += L[i][k]*A[k][j];

   return R;
}

Matrix FeaTure::PostMultLowerT( const Matrix &A, const Matrix &L )
{
   // L is a lower left triangular matrix L(i,j) j>=i
   // the upper right part is not used and need not be initialized
   //                T
   // Calculate R=A*L

   size_t n( L.size1() );
   assert( L.size2()==n && A.size2()==n );
   size_t m( A.size1() );

   Matrix R(m,n,0.0);

   for ( size_t i=0; i<m; i++ )
      for ( size_t j=0; j<n; j++ )
         for ( size_t k=0; k<=j; k++ ) R[i][j] += A[i][k]*L[j][k];

   return R;
}

Matrix FeaTure::PreMultLowerT( const Matrix &L, const Matrix &A )
{
   // L is a lower left triangular matrix L(i,j) j>=i
   // the upper right part is not used and need not be initialized
   //                T
   // Calculate R = L * A

   size_t n( L.size1() );
   assert( L.size2()==n && A.size1()==n );
   size_t m( A.size2() );

   Matrix R(n,m,0.0);

   for ( size_t i=0; i<n; i++ )
      for ( size_t j=0; j<m; j++ )
         for ( size_t k=i; k<n; k++ ) R[i][j] += L[k][i]*A[k][j];

   return R;
}

