// Copyright (c) 2003, A.H. van den Boogaard
//
// test program for eigenvalue and eigenvector analysis of
// a FeaTure::Matrix

#include <ftmatrix.h>
#include <sysmat.h>
#include <eigen.h>
#include <triangular.h>

#include <iomanip>
#include <iostream>

using FeaTure::Matrix;
using FeaTure::Vector;
using FeaTure::SparseSysMat;
using std::cout;
using std::setw;
using std::endl;

int main()
{
   cout.precision(7);

   Matrix A(3,3);
   A[0][0] = 10;
   A[0][1] = A[1][0] = 20.0;
   A[0][2] = A[2][0] = 30.0;
   A[1][1] = 45.0;
   A[1][2] = A[2][1] = 80.0;
   A[2][2] = 171;

   cout << "A" << endl
        << setw(5) << A << endl;

   Matrix U(3,3,0.0);
   for ( size_t i=0; i<3; i++ ) U[i][i]=1.0;
   Matrix AA(A);

   householder(&A,&U);

   cout << "householder(A)" << endl
        << setw(14) << A << endl;

   cout << "Ut*A*U" << endl
        << setw(14) << ~U*AA*U << endl;


   // QR steps

   A = AA;
   cout << endl << "Symmetric QR iterations" << endl;
   symmetricQR( &A, &U );

   cout << "U" << endl << setw(14) << U << endl;

   cout << "Ut*A*U" << endl
        << setw(14) << ~U*AA*U << endl;

   A = AA;

   cout << "Sturm sequence number -3: " << SturmSequence(A,-3) << endl;
   cout << "Sturm sequence number  0: " << SturmSequence(A) << endl;
   cout << "Sturm sequence number  1: " << SturmSequence(A,1) << endl;
   cout << "Sturm sequence number  2: " << SturmSequence(A,2) << endl;
   cout << "Sturm sequence number  5: " << SturmSequence(A,5) << endl;
   cout << "Sturm sequence number 15: " << SturmSequence(A,15) << endl;
   cout << endl;

   Vector v(A.size1(),0.0);
   double lambda;
   InverseIteration( A, &v, &lambda );

   cout << "inverse iteration" << endl;
   cout << "lambda " << lambda << endl;
   cout << "v " << endl << setw(15) << v << endl;

   cout << endl << "Triangular matrices" << endl;

   A = AA;
   cout << "A" << endl << setw(14) << A << endl;
   LDLtDecomposition( &A );
   cout << "LDLt" << endl << setw(14) << A << endl;
   Vector b(3);
   b[0] = 5;
   b[1] = 1;
   b[2] = 2;
   cout << "b" << endl << setw(14) << b << endl;
   LDLtSubstitution( A, &b );
   cout << "x" << endl << setw(14) << b << endl;
   cout << "Ax" << endl << setw(14) << AA*b << endl;
   
   A = AA;
   cout << "A" << endl << setw(14) << A << endl;
   CholeskyDecomposition( &A );
   cout << "LLt" << endl << setw(14) << A << endl;
   b[0] = 5;
   b[1] = 1;
   b[2] = 2;
   cout << "b" << endl << setw(14) << b << endl;
   CholeskySubstitution( A, &b );
   cout << "x" << endl << setw(14) << b << endl;
   cout << "Ax" << endl << setw(14) << AA*b << endl;

   Matrix B;

   A = AA;
   cout << "A" << endl << setw(14) << A << endl;
   CholeskyDecomposition( &A );
   cout << "L" << endl << setw(14) << A << endl;

   B = AA;
   PreMultLowerInv( A, &B );
   cout << "Linv*A" << endl << setw(14) << B << endl;
   cout << "LB" << endl << setw(14) << PreMultLower(A,B) << endl;

   B = AA;
   PostMultLowerTInv( &B, A );
   cout << "B" << endl << setw(14) << B << endl;
   cout << "BLt" << endl << setw(14) << PostMultLowerT(B,A) << endl;

   B = AA;
   PreMultLowerTInv( A, &B );
   cout << "Ltinv*A" << endl << setw(14) << B << endl;
   cout << "LtB" << endl << setw(14) << PreMultLowerT(A,B) << endl;

   Matrix K(3,3,0);
   Matrix M(3,3,0);

   K[0][0] = 1;
   K[1][0] = K[0][1] = -1;
   K[1][1] = 2;
   K[1][2] = K[2][1] = -1;
   K[2][2] = 1;

   M[0][0] = 2;
   M[1][1] = 1;
   M[2][2] = 2;

   Vector ev;

   cout << endl << "Generalized eigenproblem" << endl;
   Matrix KK(K);
   Matrix MM(M);
   Matrix V;
   GenEigen( &K, &M, &ev, &V );
   cout << "eigen values" << endl << setw(14) << ev << endl;
   cout << "eigen vectors" << endl << setw(14) << V << endl;

   cout << "~V*K*V" << endl << setw(14) << ~V*KK*V << endl;
   cout << "~V*M*V" << endl << setw(14) << ~V*MM*V << endl;

   SparseSysMat Ks(10,true);
   SparseSysMat Ms(10,true);

   for ( size_t i=1; i<10; i++ )
   {
      Ks.add( i, i, 2.0 );
      Ks.add( i+1, i, -1.0 );
      Ks.add( i, i+1, -1.0 );
      Ms.add( i, i, 1.0 );
   }
   Ks.add( 10, 10, 1.0 );
   Ms.add( 10, 10, 0.5 );
   
   size_t p(5);
   Matrix X(10,p,0.0);
   for ( size_t i=0; i<10; i++ )
      X[i][0] = Ms.get(i,i);
   for ( size_t i=1; i<p; i++ )
      X[i][i] = 1.0;

   cout << endl << "Subspace iteration" << endl;
   subspace_iteration( Ks, Ms, &ev, &X );
   cout << "eigen values" << endl << ev << endl;
   cout.precision(5);
   cout << "eigen vectors" << endl << setw(12) << X << endl;

   return 0;
}
